"""Interfaces with Amber."""

import textwrap
from collections import defaultdict
from collections.abc import Iterable
from copy import deepcopy
from pathlib import Path

import numpy
from openff.toolkit import Topology

from openff.interchange import Interchange
from openff.interchange.components.toolkit import _get_num_h_bonds
from openff.interchange.constants import (
    _PME,
    AMBER_COULOMBS_CONSTANT,
    kcal_mol,
    kcal_mol_a2,
    kcal_mol_rad2,
)
from openff.interchange.exceptions import (
    UnsupportedExportError,
    UnsupportedMixingRuleError,
)
from openff.interchange.models import PotentialKey


def _write_text_blob(file, blob):
    def chunker(seq, size):
        return (seq[pos : pos + size] for pos in range(0, len(seq), size))

    if blob == "":
        file.write("\n")
    else:
        for line in chunker(blob, 80):
            file.write(line + "\n")


def _flatten(list_of_lists: Iterable[list[int]]) -> list[int]:
    return [item for sublist in list_of_lists for item in sublist]


def _get_per_atom_exclusion_lists(
    topology: "Topology",
) -> dict[int, defaultdict[int, list[int]]]:
    """
    Get the excluded atoms of each atom in this topology.

    Parameters
    ----------
    topology
        OpenFF Topology

    Returns
    -------
    per_atom_exclusions
        keys: atom indices (OpenFF atoms, zero-indexed)
        values: defaultdict[int, list[int]]
            list of atom indices (OpenFF atoms, zero-indexed) that are excluded from this atom,
            keyed by their separation distance in the graph (1 for bonded, 2 for in an angle, 3
            for in a dihedral)

    """
    per_atom_exclusions: dict[int, defaultdict[int, list[int]]] = dict()

    for atom1 in topology.atoms:
        # Excluded atoms _on this atom_
        this_atom_exclusions = defaultdict(list)
        index1 = topology.atom_index(atom1)

        for atom2 in atom1.bonded_atoms:
            index2 = topology.atom_index(atom2)

            if index2 > index1:
                # These atoms are bonded
                this_atom_exclusions[1].append(index2)

            for atom3 in atom2.bonded_atoms:
                atom3_index = topology.atom_index(atom3)

                if atom3_index > index1:
                    # These atoms are included in an angle
                    this_atom_exclusions[2].append(atom3_index)

                for atom4 in atom3.bonded_atoms:
                    atom4_index = topology.atom_index(atom4)
                    if atom4_index > index1 and atom4_index:
                        # These atoms are included in a torsion
                        this_atom_exclusions[3].append(atom4_index)

        per_atom_exclusions[index1] = this_atom_exclusions

    return per_atom_exclusions


def _get_exclusion_lists(
    per_atom_exclusions: dict[int, defaultdict[int, list[int]]],
) -> tuple[list[int], list[int]]:
    """
    Convert a per-atom exclusion dict to Amer structures.

    Parameters
    ----------
    per_atom_exclusions
        keys: atom indices (OpenFF atoms, zero-indexed)
        values: list of atom indices (OpenFF atoms, zero-indexed) that are excluded from this atom

        See _getper_atom_exclusion_lists

    Returns
    -------
    number_excluded_atoms
        Number of excluded atoms for each atom (OpenFF atoms, zero-indexed)
    excluded_atoms_list
        Flattened list of per-atom exclusions (Amber atoms, one-indexed).
        See EXCLUDED_ATOMS_LIST in https://ambermd.org/prmtop.pdf

    """
    excluded_atoms_list: list[int] = list()
    number_excluded_atoms: list[int] = list()

    for this_atom, this_atom_exclusions in per_atom_exclusions.items():
        if len(this_atom_exclusions) == 0:
            # Amber expects this list to have a 1 in it, pointing to a non-existent
            # atom with index 0, when an atom has no exclusions
            tmp = {0}

        else:
            tmp = set()
            for other_atom in _flatten(this_atom_exclusions.values()):
                if other_atom > this_atom:
                    tmp.add(other_atom + 1)

        for val in sorted(tmp):
            excluded_atoms_list.append(val)

        number_excluded_atoms.append(len(tmp))

    return number_excluded_atoms, excluded_atoms_list


def _get_bond_lists(
    interchange: "Interchange",
    atomic_numbers: tuple,
    potential_key_to_bond_type_mapping: dict[PotentialKey, int],
) -> tuple[list[int], list[int]]:
    bonds_inc_hydrogen: list[int] = list()
    bonds_without_hydrogen: list[int] = list()

    # TODO: Should probably build bond lists and exclusions without assuming bond physics
    for bond, key_ in interchange["Bonds"].key_map.items():
        bond_type_index = potential_key_to_bond_type_mapping[key_]

        these_atomic_numbers = tuple(atomic_numbers[index] for index in bond.atom_indices)

        bond_indices = tuple(sorted(bond.atom_indices))

        bonds_list = bonds_inc_hydrogen if 1 in these_atomic_numbers else bonds_without_hydrogen

        bonds_list.append(bond_indices[0] * 3)
        bonds_list.append(bond_indices[1] * 3)
        bonds_list.append(bond_type_index + 1)

    return bonds_inc_hydrogen, bonds_without_hydrogen


def _get_angle_lists(
    interchange: "Interchange",
    atomic_numbers: tuple,
    potential_key_to_angle_type_mapping: dict[PotentialKey, int],
) -> tuple[list[int], list[int]]:
    angles_inc_hydrogen: list[int] = list()
    angles_without_hydrogen: list[int] = list()

    for angle, key_ in interchange["Angles"].key_map.items():
        angle_type_index = potential_key_to_angle_type_mapping[key_]

        these_atomic_numbers = tuple(atomic_numbers[index] for index in angle.atom_indices)

        angle_indices = list(angle.atom_indices)
        angle_indices = angle_indices if angle_indices[0] < angle_indices[-1] else list(reversed(angle_indices))

        angles_list = angles_inc_hydrogen if 1 in these_atomic_numbers else angles_without_hydrogen

        angles_list.append(angle_indices[0] * 3)
        angles_list.append(angle_indices[1] * 3)
        angles_list.append(angle_indices[2] * 3)
        angles_list.append(angle_type_index + 1)

    return angles_inc_hydrogen, angles_without_hydrogen


def _get_dihedral_lists(
    interchange: "Interchange",
    atomic_numbers: tuple,
    potential_key_to_dihedral_type_mapping: dict[PotentialKey, int],
    already_counted: set[tuple[int, ...]],
    exclusions: dict[int, defaultdict[int, list[int]]],  # per-atom exclusions
) -> tuple[list[int], list[int]]:
    dihedrals_inc_hydrogen: list[int] = list()
    dihedrals_without_hydrogen: list[int] = list()

    if "ProperTorsions" in interchange.collections:
        for dihedral, proper_key in interchange["ProperTorsions"].key_map.items():
            dihedral_type_index = potential_key_to_dihedral_type_mapping[proper_key]

            atom1_index, atom2_index, atom3_index, atom4_index = dihedral.atom_indices

            # Since 0 can't be negative, attempt to re-arrange this torsion
            # such that the third atom listed is negative.
            # This should only be strictly necessary when _14_tag is -1, but
            # ParmEd likes to always flip it, and always flipping should be harmless.
            # Could put this in an if block if desired.
            if atom3_index == 0:
                atom1_index, atom2_index, atom3_index, atom4_index = reversed(
                    (
                        atom1_index,
                        atom2_index,
                        atom3_index,
                        atom4_index,
                    ),
                )

            these_atomic_numbers = tuple(atomic_numbers[index] for index in dihedral.atom_indices)

            # See if the non-bonded interactions of the first and fourth atoms need to be exluded

            atom1_neighbors = exclusions[atom1_index][1] + exclusions[atom1_index][2]
            atom4_neighbors = exclusions[atom4_index][1] + exclusions[atom4_index][2]

            _sorted = tuple(sorted([atom1_index, atom4_index]))

            if atom4_index in atom1_neighbors:
                # Exclude because these atoms are separated by only one or two bonds
                exclude_nonbonded = -1

            elif atom1_index in atom4_neighbors:
                # Exclude because these atoms are separated by only one or two bonds
                exclude_nonbonded = -1

            elif _sorted in already_counted:
                # Exclude because these were already counted once in another torsion
                exclude_nonbonded = -1

            else:
                exclude_nonbonded = 1

            dihedrals_list = dihedrals_inc_hydrogen if 1 in these_atomic_numbers else dihedrals_without_hydrogen

            dihedrals_list.append(atom1_index * 3)
            dihedrals_list.append(atom2_index * 3)
            dihedrals_list.append(atom3_index * 3 * exclude_nonbonded)
            dihedrals_list.append(atom4_index * 3)
            dihedrals_list.append(dihedral_type_index + 1)

            already_counted.add(_sorted)

    if "ImproperTorsions" in interchange.collections:
        for improper, improper_key in interchange["ImproperTorsions"].key_map.items():
            dihedral_type_index = potential_key_to_dihedral_type_mapping[improper_key]

            atom1_index, atom2_index, atom3_index, atom4_index = improper.atom_indices

            # Assume that no improper torsions include 1-4 pairs, so don't check nor track them

            these_atomic_numbers = tuple(atomic_numbers[index] for index in improper.atom_indices)

            dihedrals_list = dihedrals_inc_hydrogen if 1 in these_atomic_numbers else dihedrals_without_hydrogen

            dihedrals_list.append(atom1_index * 3)
            dihedrals_list.append(atom2_index * 3)
            dihedrals_list.append(atom3_index * 3 * -1)
            dihedrals_list.append(atom4_index * 3 * -1)
            dihedrals_list.append(dihedral_type_index + 1)

    return dihedrals_inc_hydrogen, dihedrals_without_hydrogen


# TODO: Split this mono-function into smaller functions
def to_prmtop(interchange: "Interchange", file_path: Path | str):
    """
    Write a .prmtop file. See http://ambermd.org/prmtop.pdf for details.

    """
    if isinstance(file_path, str):
        path = Path(file_path)
    if isinstance(file_path, Path):
        path = file_path

    if interchange["vdW"].mixing_rule != "lorentz-berthelot":
        raise UnsupportedMixingRuleError(interchange["vdW"].mixing_rule)

    if "VirtualSites" in interchange.collections:
        if len(interchange["VirtualSites"].key_map) > 0:
            raise UnsupportedExportError(
                "Virtual sites are not yet supported in Amber writers. See issues labeled "
                "'amber' or 'virtual sites' for status and updates of this feature.",
            )

    if interchange.box is None:
        if interchange["Electrostatics"].periodic_potential != _PME:
            raise UnsupportedExportError(
                f'Electrostatics method PME (`"{_PME}"`) is not valid for a non-periodic system. ',
            )

    with open(path, "w") as prmtop:
        import datetime

        now = datetime.datetime.now()
        prmtop.write(
            "%VERSION  VERSION_STAMP = V0001.000  DATE = "
            f"{now.month:02d}/{now.day:02d}/{(now.year % 100):02}  "
            f"{now.hour:02d}:{now.minute:02d}:{now.second:02d}\n"
            "%FLAG TITLE\n"
            "%FORMAT(20a4)\n"
            "\n",
        )

        from openff.interchange.interop.common import _build_typemap

        typemap = _build_typemap(interchange)

        potential_key_to_atom_type_mapping: dict[PotentialKey, int] = {
            key: i for i, key in enumerate(interchange["vdW"].potentials)
        }
        atom_type_indices = [
            potential_key_to_atom_type_mapping[potential_key] for potential_key in interchange["vdW"].key_map.values()
        ]

        potential_key_to_bond_type_mapping: dict[PotentialKey, int] = {
            key: i for i, key in enumerate(interchange["Bonds"].potentials)
        }

        potential_key_to_angle_type_mapping: dict[PotentialKey, int] = {
            key: i for i, key in enumerate(interchange["Angles"].potentials)
        }

        dihedral_potentials = dict()
        for key in ["ProperTorsions", "ImproperTorsions"]:
            if key in interchange.collections:
                dihedral_potentials.update(deepcopy(interchange[key].potentials))
                dihedral_potentials.update(deepcopy(interchange[key].potentials))

        potential_key_to_dihedral_type_mapping: dict[PotentialKey, int] = {
            key: i for i, key in enumerate(dihedral_potentials)
        }

        # This is only used to track 1-4 pairs that are already counted. Since we
        # only care about membership and never indexing and the membership check
        # will happen roughly once per dihedral, use a set for speed.
        already_counted: set[tuple[int, ...]] = set()

        per_atom_exclusion_lists = _get_per_atom_exclusion_lists(interchange.topology)

        number_excluded_atoms, excluded_atoms_list = _get_exclusion_lists(
            per_atom_exclusion_lists,
        )
        atomic_numbers = tuple(atom.atomic_number for atom in interchange.topology.atoms)

        bonds_inc_hydrogen, bonds_without_hydrogen = _get_bond_lists(
            interchange,
            atomic_numbers,
            potential_key_to_bond_type_mapping,
        )

        angles_inc_hydrogen, angles_without_hydrogen = _get_angle_lists(
            interchange,
            atomic_numbers,
            potential_key_to_angle_type_mapping,
        )

        dihedrals_inc_hydrogen, dihedrals_without_hydrogen = _get_dihedral_lists(
            interchange,
            atomic_numbers,
            potential_key_to_dihedral_type_mapping,
            already_counted,
            per_atom_exclusion_lists,
        )

        # total number of atoms
        NATOM = interchange.topology.n_atoms
        # total number of distinct atom types
        NTYPES = len(interchange["vdW"].potentials)
        # number of bonds containing hydrogen
        NBONH = _get_num_h_bonds(interchange.topology)
        # number of bonds not containing hydrogen
        MBONA = interchange.topology.n_bonds - NBONH
        # number of angles containing hydrogen
        NTHETH = int(len(angles_inc_hydrogen) / 4)
        # number of angles not containing hydrogen
        MTHETA = int(len(angles_without_hydrogen) / 4)
        # number of dihedrals containing hydrogen
        NPHIH = int(len(dihedrals_inc_hydrogen) / 5)
        # number of dihedrals not containing hydrogen
        MPHIA = int(len(dihedrals_without_hydrogen) / 5)
        NHPARM = 0  # : currently not used
        NPARM = 0  # : used to determine if addles created prmtop
        # number of excluded atoms
        NNB = len(excluded_atoms_list)
        # number of residues
        NRES = max(1, len([*interchange.topology.hierarchy_iterator("residues")]))
        NBONA = MBONA  # : MBONA + number of constraint bonds
        NTHETA = MTHETA  # : MTHETA + number of constraint angles
        NPHIA = MPHIA  # : MPHIA + number of constraint dihedrals
        # number of unique bond types
        NUMBND = len(potential_key_to_bond_type_mapping)
        # number of unique angle types
        NUMANG = len(potential_key_to_angle_type_mapping)
        # number of unique dihedral types
        NPTRA = len(potential_key_to_dihedral_type_mapping)
        # number of atom types in parameter file, see SOLTY below
        # this appears to be unused, but ParmEd writes a 1 here (?)
        NATYP = 1
        NPHB = 0  # : number of distinct 10-12 hydrogen bond pair types
        IFPERT = 0  # : set to 1 if perturbation info is to be read in
        NBPER = 0  # : number of bonds to be perturbed
        NGPER = 0  # : number of angles to be perturbed
        NDPER = 0  # : number of dihedrals to be perturbed
        MBPER = 0  # : number of bonds with atoms completely in perturbed group
        MGPER = 0  # : number of angles with atoms completely in perturbed group
        MDPER = 0  # : number of dihedrals with atoms completely in perturbed groups
        # set to 1 if standard periodic box, 2 when truncated octahedral
        IFBOX = 0 if interchange.box is None else 1
        # number of atoms in the largest residue
        NMXRS = interchange.topology.n_atoms
        IFCAP = 0  # : set to 1 if the CAP option from edit was specified
        NUMEXTRA = 0  # : number of extra points found in topology
        # number of PIMD slices / number of beads
        # ParmEd does not seem to write this _at all_ in most/all cases
        NCOPY = 0

        pointers = [
            NATOM,
            NTYPES,
            NBONH,
            MBONA,
            NTHETH,
            MTHETA,
            NPHIH,
            MPHIA,
            NHPARM,
            NPARM,
            NNB,
            NRES,
            NBONA,
            NTHETA,
            NPHIA,
            NUMBND,
            NUMANG,
            NPTRA,
            NATYP,
            NPHB,
            IFPERT,
            NBPER,
            NGPER,
            NDPER,
            MBPER,
            MGPER,
            MDPER,
            IFBOX,
            NMXRS,
            IFCAP,
            NUMEXTRA,
            NCOPY,
        ]

        prmtop.write("%FLAG POINTERS\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in pointers])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG ATOM_NAME\n%FORMAT(20a4)\n")

        atom_names: list[str] = list()

        for atom in interchange.topology.atoms:
            if hasattr(atom, "name"):
                if atom.name:
                    atom_names.append(atom.name)
                else:
                    atom_names.append(atom.symbol)
            else:
                atom_names.append(atom.symbol)

        text_blob = "".join([str(val).rjust(4) for val in atom_names])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG CHARGE\n%FORMAT(5E16.8)\n")
        charges = (
            interchange["Electrostatics"]
            .get_charge_array()
            .m_as(
                "elementary_charge",
            )
            * AMBER_COULOMBS_CONSTANT
        ).tolist()
        text_blob = "".join([f"{val:16.8E}" for val in charges])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG ATOMIC_NUMBER\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in atomic_numbers])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG MASS\n%FORMAT(5E16.8)\n")
        masses = [a.mass.m for a in interchange.topology.atoms]
        text_blob = "".join([f"{val:16.8E}" for val in masses])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG ATOM_TYPE_INDEX\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val + 1).rjust(8) for val in atom_type_indices])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG NUMBER_EXCLUDED_ATOMS\n%FORMAT(10I8)\n")
        # https://ambermd.org/prmtop.pdf says this section is ignored (!?)
        text_blob = "".join([str(val).rjust(8) for val in number_excluded_atoms])
        _write_text_blob(prmtop, text_blob)

        acoefs = [None] * int((NTYPES + 1) * NTYPES / 2)
        bcoefs = [None] * int((NTYPES + 1) * NTYPES / 2)

        nonbonded_parm_indices: list[int | None] = [None] * (NTYPES * NTYPES)

        for key_i, i in potential_key_to_atom_type_mapping.items():
            for key_j, j in potential_key_to_atom_type_mapping.items():
                if j < i:
                    # Only need to handle the lower triangle as everything symmetric.
                    continue

                atom_type_index_i = i + 1
                atom_type_index_j = j + 1

                coeff_index = int(i + (j + 1) * j / 2) + 1  # FORTRAN IDX

                # index = NONBONDED PARM INDEX [NTYPES * (ATOM TYPE INDEX(i) - 1) + ATOM TYPE INDEX(j)]
                parm_index_fwd = (
                    NTYPES * (atom_type_index_i - 1) + atom_type_index_j  # FORTRAN IDX
                )
                parm_index_rev = (
                    NTYPES * (atom_type_index_j - 1) + atom_type_index_i  # FORTRAN IDX
                )

                # TODO: Figure out the right way to map cross-interactions, using the
                #       key_i and key_j objects as lookups to parameters
                sigma_i = interchange["vdW"].potentials[key_i].parameters["sigma"]
                sigma_j = interchange["vdW"].potentials[key_j].parameters["sigma"]
                epsilon_i = interchange["vdW"].potentials[key_i].parameters["epsilon"]
                epsilon_j = interchange["vdW"].potentials[key_j].parameters["epsilon"]

                sigma = (sigma_i + sigma_j) * 0.5
                epsilon = (epsilon_i * epsilon_j) ** 0.5

                acoef = (4 * epsilon * sigma**12).m_as("kcal / mol * angstrom**12")
                bcoef = (4 * epsilon * sigma**6).m_as("kcal / mol * angstrom**6")

                acoefs[coeff_index - 1] = acoef
                bcoefs[coeff_index - 1] = bcoef

                nonbonded_parm_indices[parm_index_fwd - 1] = coeff_index
                nonbonded_parm_indices[parm_index_rev - 1] = coeff_index

        assert all(
            value is not None
            for values in [acoefs, bcoefs, nonbonded_parm_indices]
            for value in values  # type: ignore
        ), "an internal error occurred"

        prmtop.write("%FLAG NONBONDED_PARM_INDEX\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in nonbonded_parm_indices])
        _write_text_blob(prmtop, text_blob)

        residue_names = [
            getattr(residue, "residue_name", "RES") for residue in interchange.topology.hierarchy_iterator("residues")
        ]

        # Avoid blank residue labels, though this is a band-aid. See issue #652
        if residue_names == list():
            residue_names = ["RES"]

        prmtop.write("%FLAG RESIDUE_LABEL\n%FORMAT(20a4)\n")
        text_blob = "".join([val.ljust(4) for val in residue_names])
        _write_text_blob(prmtop, text_blob)

        residue_pointers = (
            [
                interchange.topology.atom_index(next(iter(residue.atoms)))
                for residue in interchange.topology.hierarchy_iterator("residues")
            ]
            if NRES > 1
            else [0]
        )
        prmtop.write("%FLAG RESIDUE_POINTER\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val + 1).rjust(8) for val in residue_pointers])
        _write_text_blob(prmtop, text_blob)

        # TODO: Exclude (?) bonds containing hydrogens
        prmtop.write("%FLAG BOND_FORCE_CONSTANT\n%FORMAT(5E16.8)\n")
        bond_k = [
            interchange["Bonds"].potentials[key].parameters["k"].m_as(kcal_mol_a2) / 2
            for key in potential_key_to_bond_type_mapping
        ]
        text_blob = "".join([f"{val:16.8E}" for val in bond_k])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG BOND_EQUIL_VALUE\n%FORMAT(5E16.8)\n")
        bond_length = [
            interchange["Bonds"].potentials[key].parameters["length"].m_as("angstrom")
            for key in potential_key_to_bond_type_mapping
        ]
        text_blob = "".join([f"{val:16.8E}" for val in bond_length])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG ANGLE_FORCE_CONSTANT\n%FORMAT(5E16.8)\n")
        angle_k = [
            interchange["Angles"].potentials[key].parameters["k"].m_as(kcal_mol_rad2) / 2
            for key in potential_key_to_angle_type_mapping
        ]
        text_blob = "".join([f"{val:16.8E}" for val in angle_k])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG ANGLE_EQUIL_VALUE\n%FORMAT(5E16.8)\n")
        angle_theta = [
            interchange["Angles"].potentials[key].parameters["angle"].m_as("radian")
            for key in potential_key_to_angle_type_mapping
        ]
        text_blob = "".join([f"{val:16.8E}" for val in angle_theta])
        _write_text_blob(prmtop, text_blob)

        dihedral_k: list[int] = list()
        dihedral_periodicity: list[int] = list()
        dihedral_phase: list[int] = list()

        for key_ in potential_key_to_dihedral_type_mapping:
            params = interchange[key_.associated_handler].potentials[key_].parameters  # type: ignore
            idivf = int(params["idivf"]) if "idivf" in params else 1
            dihedral_k.append((params["k"] / idivf).m_as(kcal_mol))
            dihedral_periodicity.append(params["periodicity"].m_as("dimensionless"))
            dihedral_phase.append(params["phase"].m_as("radian"))

        prmtop.write("%FLAG DIHEDRAL_FORCE_CONSTANT\n%FORMAT(5E16.8)\n")
        text_blob = "".join([f"{val:16.8E}" for val in dihedral_k])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG DIHEDRAL_PERIODICITY\n%FORMAT(5E16.8)\n")
        text_blob = "".join([f"{val:16.8E}" for val in dihedral_periodicity])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG DIHEDRAL_PHASE\n%FORMAT(5E16.8)\n")
        text_blob = "".join([f"{val:16.8E}" for val in dihedral_phase])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG SCEE_SCALE_FACTOR\n%FORMAT(5E16.8)\n")
        scee = NPTRA * [1.2]
        text_blob = "".join([f"{val:16.8E}" for val in scee])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG SCNB_SCALE_FACTOR\n%FORMAT(5E16.8)\n")
        scnb = NPTRA * [2.0]
        text_blob = "".join([f"{val:16.8E}" for val in scnb])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG SOLTY\n%FORMAT(5E16.8)\n")
        prmtop.write(f"{0:16.8E}\n")

        prmtop.write("%FLAG LENNARD_JONES_ACOEF\n%FORMAT(5E16.8)\n")
        text_blob = "".join([f"{val:16.8E}" for val in acoefs])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG LENNARD_JONES_BCOEF\n%FORMAT(5E16.8)\n")
        text_blob = "".join([f"{val:16.8E}" for val in bcoefs])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG BONDS_INC_HYDROGEN\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in bonds_inc_hydrogen])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG BONDS_WITHOUT_HYDROGEN\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in bonds_without_hydrogen])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG ANGLES_INC_HYDROGEN\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in angles_inc_hydrogen])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG ANGLES_WITHOUT_HYDROGEN\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in angles_without_hydrogen])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG DIHEDRALS_INC_HYDROGEN\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in dihedrals_inc_hydrogen])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG DIHEDRALS_WITHOUT_HYDROGEN\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in dihedrals_without_hydrogen])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG EXCLUDED_ATOMS_LIST\n%FORMAT(10I8)\n")
        text_blob = "".join([str(val).rjust(8) for val in excluded_atoms_list])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG HBOND_ACOEF\n%FORMAT(5E16.8)\n")
        _write_text_blob(prmtop, "")

        prmtop.write("%FLAG HBOND_BCOEF\n%FORMAT(5E16.8)\n")
        _write_text_blob(prmtop, "")

        prmtop.write("%FLAG HBCUT\n%FORMAT(5E16.8)\n")
        _write_text_blob(prmtop, "")

        prmtop.write("%FLAG AMBER_ATOM_TYPE\n%FORMAT(20a4)\n")
        text_blob = "".join([val.ljust(4) for val in typemap.values()])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG TREE_CHAIN_CLASSIFICATION\n%FORMAT(20a4)\n")
        blahs = NATOM * ["BLA"]
        text_blob = "".join([val.ljust(4) for val in blahs])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG JOIN_ARRAY\n%FORMAT(10I8)\n")
        _ = NATOM * [0]
        text_blob = "".join([str(val).rjust(8) for val in _])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG IROTAT\n%FORMAT(10I8)\n")
        _ = NATOM * [0]
        text_blob = "".join([str(val).rjust(8) for val in _])
        _write_text_blob(prmtop, text_blob)

        if IFBOX == 1:
            if (interchange.box.m != numpy.diag(numpy.diagonal(interchange.box.m))).any():  # type: ignore[union-attr]
                raise NotImplementedError(
                    "Interchange does not yet support exporting non-rectangular boxes to Amber",
                )

            prmtop.write("%FLAG SOLVENT_POINTERS\n%FORMAT(3I8)\n")
            prmtop.write("       1       1       2\n")

            # TODO: No easy way to accurately export this section while
            #       using an MDTraj topology
            prmtop.write("%FLAG ATOMS_PER_MOLECULE\n%FORMAT(10I8)\n")
            prmtop.write(str(interchange.topology.n_atoms).rjust(8))
            prmtop.write("\n")

            prmtop.write("%FLAG BOX_DIMENSIONS\n%FORMAT(5E16.8)\n")
            box = [90.0]
            for i in range(3):
                box.append(interchange.box[i, i].m_as("angstrom"))  # type: ignore
            text_blob = "".join([f"{val:16.8E}" for val in box])
            _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG RADIUS_SET\n%FORMAT(1a80)\n")
        prmtop.write("0\n")

        prmtop.write("%FLAG RADII\n%FORMAT(5E16.8)\n")
        radii = NATOM * [0]
        text_blob = "".join([f"{val:16.8E}" for val in radii])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG SCREEN\n%FORMAT(5E16.8)\n")
        screen = NATOM * [0]
        text_blob = "".join([f"{val:16.8E}" for val in screen])
        _write_text_blob(prmtop, text_blob)

        prmtop.write("%FLAG IPOL\n%FORMAT(1I8)\n")
        prmtop.write("       0\n")


def to_inpcrd(interchange: "Interchange", file_path: Path | str):
    """
    Write a .prmtop file. See https://ambermd.org/FileFormats.php#restart for details.

    """
    if isinstance(file_path, str):
        path = Path(file_path)
    if isinstance(file_path, Path):
        path = file_path

    n_atoms = interchange.topology.n_atoms
    time = 0.0

    with open(path, "w") as inpcrd:
        inpcrd.write(f"\n{n_atoms:5d}{time:15.7e}\n")

        coords = interchange.positions.m_as("angstrom")  # type: ignore[union-attr]
        blob = "".join([f"{val:12.7f}".rjust(12) for val in coords.flatten()])

        for line in textwrap.wrap(blob, width=72, drop_whitespace=False):
            inpcrd.write(line + "\n")

        if interchange.box is not None:
            box = interchange.box.to("angstrom").magnitude
            if (box == numpy.diag(numpy.diagonal(box))).all():
                for i in range(3):
                    inpcrd.write(f"{box[i, i]:12.7f}")
                for _ in range(3):
                    inpcrd.write("  90.0000000")
            else:
                # TODO: Handle non-rectangular
                raise NotImplementedError(
                    "Interchange does not yet support exporting non-rectangular boxes to Amber",
                )

        inpcrd.write("\n")



import pytest
from openff.toolkit import Molecule, Quantity
from openff.toolkit.typing.engines.smirnoff.parameters import BondHandler

from openff.interchange.components.potentials import (
    Collection,
    Potential,
    WrappedPotential,
)
from openff.interchange.smirnoff._valence import SMIRNOFFBondCollection


class TestWrappedPotential:
    def test_interpolated_potentials(self):
        """Test the construction of and .parameters getter of WrappedPotential"""

        bt = BondHandler.BondType(
            smirks="[#6X4:1]~[#8X2:2]",
            id="bbo1",
            k_bondorder1="100.0 * kilocalories_per_mole/angstrom**2",
            k_bondorder2="200.0 * kilocalories_per_mole/angstrom**2",
            length_bondorder1="1.4 * angstrom",
            length_bondorder2="1.3 * angstrom",
        )

        pot1 = Potential(
            parameters={"k": bt.k_bondorder[1], "length": bt.length_bondorder[1]},
        )
        pot2 = Potential(
            parameters={"k": bt.k_bondorder[2], "length": bt.length_bondorder[2]},
        )

        interp_pot = WrappedPotential(data={pot1: 0.2, pot2: 0.8})
        assert interp_pot.parameters == {
            "k": Quantity(180, "kilocalorie / angstrom ** 2 / mole"),
            "length": Quantity(1.32, "angstrom"),
        }

        # Ensure a single Potential object can be wrapped with similar behavior
        simple = WrappedPotential(data=pot2)
        assert simple.parameters == pot2.parameters


class TestCosmeticAttributes:
    def test_potential_with_cosmetic_attributes(self):
        cosmetic = BondHandler.BondType(
            smirks="[*:1]~[*:2]",
            id="cos1",
            k="430.0 * kilocalories_per_mole/angstrom**2",
            length="1.33 * angstrom",
            foo="bar",
            allow_cosmetic_attributes=True,
        )

        handler = BondHandler(version=0.4)
        handler.add_parameter(parameter=cosmetic)

        collection = SMIRNOFFBondCollection()

        collection.store_matches(
            parameter_handler=handler,
            topology=Molecule.from_smiles("O").to_topology(),
        )


class TestCollectionSubclassing:
    def test_dummy_collection(self):
        handler = Collection(
            type="foo",
            expression="m*x+b",
        )
        assert handler.type == "foo"
        assert handler.expression == "m*x+b"


class TestPotentialSerialization:
    @pytest.fixture
    def dummy_potential(self):
        return Potential(
            parameters={
                "a": Quantity(1.0, "kilocalorie / mole"),
                "b": Quantity(2.0, "angstrom"),
            },
        )

    def test_json_roundtrip(self, dummy_potential):
        potential = Potential.model_validate_json(dummy_potential.model_dump_json())

        assert potential.parameters == dummy_potential.parameters



"""An object for storing, manipulating, and converting molecular mechanics data."""

import pathlib
import tempfile
import warnings
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Union, overload

from openff.toolkit import Molecule, Quantity, Topology
from openff.utilities.utilities import has_package, requires_package
from pydantic import Field

from openff.interchange._annotations import (
    PositiveFloat,
    _BoxQuantity,
    _PositionsQuantity,
    _VelocityQuantity,
)
from openff.interchange._experimental import experimental
from openff.interchange.common._nonbonded import ElectrostaticsCollection, vdWCollection
from openff.interchange.common._valence import (
    AngleCollection,
    BondCollection,
    ImproperTorsionCollection,
    ProperTorsionCollection,
)
from openff.interchange.components.mdconfig import MDConfig
from openff.interchange.components.potentials import Collection, _AnnotatedCollections
from openff.interchange.exceptions import (
    InvalidPositionsError,
    MissingParameterHandlerError,
    MissingPositionsError,
    UnsupportedExportError,
)
from openff.interchange.operations.minimize import (
    _DEFAULT_ENERGY_MINIMIZATION_TOLERANCE,
)
from openff.interchange.pydantic import _BaseModel
from openff.interchange.serialization import _AnnotatedTopology
from openff.interchange.smirnoff import (
    SMIRNOFFConstraintCollection,
    SMIRNOFFVirtualSiteCollection,
)
from openff.interchange.smirnoff._gbsa import SMIRNOFFGBSACollection
from openff.interchange.warnings import InterchangeDeprecationWarning

if TYPE_CHECKING:
    import openmm
    import openmm.app
    from openff.toolkit import ForceField

    from openff.interchange.foyer._guard import has_foyer

    if has_foyer:
        try:
            from foyer import Forcefield as FoyerForcefield
        except ModuleNotFoundError:
            # case of openff/interchange/foyer/ being detected as the real package
            pass
    if has_package("nglview"):
        import nglview


class Interchange(_BaseModel):
    """
    A object for storing, manipulating, and converting molecular mechanics data.

    Examples
    --------
    Create an ``Interchange`` from an OpenFF ``ForceField`` and ``Molecule``

    >>> from openff.toolkit import ForceField, Molecule
    >>> sage = ForceField("openff-2.2.0.offxml")
    >>> top = Molecule.from_smiles("CCC").to_topology()
    >>> interchange = sage.create_interchange(top)

    Get the parameters for the bond between atoms 0 and 1

    >>> interchange["Bonds"][0, 1]
    Potential(...)

    """

    collections: _AnnotatedCollections = Field(dict())
    topology: _AnnotatedTopology
    mdconfig: MDConfig | None = Field(None)
    box: _BoxQuantity | None = Field(None)  # Needs shape/OpenMM validation
    positions: _PositionsQuantity | None = Field(None)  # Ditto
    velocities: _VelocityQuantity | None = Field(None)  # Ditto

    @classmethod
    def from_smirnoff(
        cls,
        force_field: "ForceField",
        topology: Topology | list[Molecule],
        box: Quantity | None = None,
        positions: Quantity | None = None,
        charge_from_molecules: list[Molecule] | None = None,
        partial_bond_orders_from_molecules: list[Molecule] | None = None,
        allow_nonintegral_charges: bool = False,
    ) -> "Interchange":
        """
        Create a new object by parameterizing a topology with a SMIRNOFF force field.

        Parameters
        ----------
        force_field
            The force field to parameterize the topology with.
        topology
            The topology to parameterize, or a list of molecules to construct a
            topology from and parameterize.
        box
            The box vectors associated with the ``Interchange``. If ``None``,
            box vectors are taken from the topology, if present.
        positions
            The positions associated with atoms in the input topology. If ``None``,
            positions are taken from the molecules in topology, if present on all molecules.
        charge_from_molecules
            If specified, partial charges for any molecules isomorphic to those
            given will be taken from the given molecules' `partial_charges`
            attribute instead of being determined by the force field. All
            molecules in this list must have partial charges assigned and must
            not be isomorphic with any other molecules in the list. For all values
            of this argument, charges on the input topology are ignored.
        partial_bond_orders_from_molecules
            If specified, partial bond orders will be taken from the given molecules
            instead of being determined by the force field.
        allow_nonintegral_charges
            If True, allow molecules to have approximately non-integral charges.

        Notes
        -----
        If the ``Molecule`` objects in the ``topology`` argument each contain
        conformers, the returned ``Interchange`` object will have its positions
        set via concatenating the 0th conformer of each ``Molecule``.

        If the ``Molecule`` objects in the ``topology`` argument have stored
        partial charges, these are ignored and charges are assigned according to
        the contents of the force field. To override the force field and use
        preset charges, use the ``charge_from_molecules`` argument.

        Examples
        --------
        Generate an Interchange object from a single-molecule (OpenFF) topology and
        OpenFF 2.0.0 "Sage"

        .. code-block:: pycon

            >>> from openff.toolkit import ForceField, Molecule
            >>> mol = Molecule.from_smiles("CC")
            >>> mol.generate_conformers(n_conformers=1)
            >>> sage = ForceField("openff-2.0.0.offxml")
            >>> interchange = sage.create_interchange(mol.to_topology())
            >>> interchange
            Interchange with 7 collections, non-periodic topology with 8 atoms.

        """
        from openff.interchange.smirnoff._create import _create_interchange

        return _create_interchange(
            force_field=force_field,
            topology=topology,
            box=box,
            positions=positions,
            molecules_with_preset_charges=charge_from_molecules,
            partial_bond_orders_from_molecules=partial_bond_orders_from_molecules,
            allow_nonintegral_charges=allow_nonintegral_charges,
        )

    def visualize(
        self,
        backend: str = "nglview",
        include_virtual_sites: bool = False,
    ) -> "nglview.NGLWidget":
        """
        Visualize this Interchange.

        This currently only uses NGLview. Other engines may be added in the future.

        Parameters
        ----------
        backend
            The backend to use for visualization. Currently only "nglview" is supported.
        include_virtual_sites
            Whether or not to include virtual sites in the visualization.

        Returns
        -------
        widget
            The NGLWidget containing the visualization.

        """
        from openff.toolkit.utils.exceptions import (
            IncompatibleUnitError,
            MissingConformersError,
        )

        if backend == "nglview":
            if include_virtual_sites:
                return self._visualize_nglview(include_virtual_sites=True)

            else:
                viz_topology = Topology(self.topology)

                try:
                    viz_topology.set_positions(self.positions)
                    return viz_topology.visualize()
                except (
                    MissingConformersError,
                    IncompatibleUnitError,
                    ValueError,
                ) as error:
                    raise MissingPositionsError(
                        "Cannot visualize system without positions.",
                    ) from error

        else:
            raise UnsupportedExportError

    @requires_package("nglview")
    def _visualize_nglview(
        self,
        include_virtual_sites: bool = False,
    ) -> "nglview.NGLWidget":
        """
        Visualize the system using NGLView via a PDB file.

        include_virtual_sites
            Whether or not to include virtual sites in the visualization.
        """
        import nglview

        from openff.interchange.components._viz import InterchangeNGLViewStructure

        try:
            widget = nglview.NGLWidget(
                InterchangeNGLViewStructure(
                    interchange=self,
                    ext="pdb",
                ),
                representations=[
                    dict(type="unitcell", params=dict()),
                ],
            )

        except MissingPositionsError as error:
            raise MissingPositionsError(
                "Cannot visualize system without positions.",
            ) from error

        widget.add_representation("line", selection="water")
        widget.add_representation("spacefill", selection="ion")
        widget.add_representation("cartoon", selection="protein")
        widget.add_representation(
            "licorice",
            selection="not water and not ion and not protein",
            radius=0.25,
            multipleBond=False,
        )

        return widget

    def minimize(
        self,
        engine: str = "openmm",
        force_tolerance: Quantity = _DEFAULT_ENERGY_MINIMIZATION_TOLERANCE,
        max_iterations: int = 10_000,
    ):
        """
        Minimize the energy of the system using an available engine.

        Updates positions in-place.

        Parameters
        ----------
        engine
            The engine to use for minimization. Currently only "openmm" is supported.
        force_tolerance
            The force tolerance to run until during energy minimization.
        max_iterations
            The maximum number of iterations to run during energy minimization.

        """
        if engine == "openmm":
            from openff.interchange.operations.minimize.openmm import minimize_openmm

            minimized_positions = minimize_openmm(
                self,
                tolerance=force_tolerance,
                max_iterations=max_iterations,
            )
            self.positions = minimized_positions
        else:
            raise NotImplementedError(f"Engine {engine} is not implemented.")

    def get_positions(self, include_virtual_sites: bool = True) -> Quantity:
        """
        Get the positions associated with this Interchange.

        Parameters
        ----------
        include_virtual_sites
            Include virtual sites in the returned positions.

        Returns
        -------
        positions
            The positions of the atoms in the system.

        """
        from openff.interchange.interop.common import _to_positions

        return _to_positions(self, include_virtual_sites=include_virtual_sites)

    def to_gromacs(
        self,
        prefix: str,
        decimal: int = 3,
        hydrogen_mass: PositiveFloat = 1.007947,
        monolithic: bool = True,
        _merge_atom_types: bool = False,
    ):
        """
        Export this Interchange object to GROMACS files.

        Parameters
        ----------
        prefix
            The prefix to use for the GROMACS topology and coordinate files, i.e. "foo" will produce
            "foo.top", "foo.gro", and "foo_pointenergy.mdp".
        decimal
            The number of decimal places to use when writing the GROMACS coordinate file.
        hydrogen_mass
            The mass to use for hydrogen atoms if not present in the topology. If non-trivially different
            than the default value, mass will be transferred from neighboring heavy atoms. Note that this is currently
            not applied to any waters and is unsupported when virtual sites are present.
        monolithic
            Whether the topology file should be monolithic (True) or reference individual .itp files (False). Note that
            these individual .itp files rely on ad hoc atom types and cannot be transferred between systems.
        _merge_atom_types
            The flag to define behaviour of GROMACSWriter. If True, then similar atom types will be merged.
            If False, each atom will have its own atom type.

        Notes
        -----
        Molecule names in written files are not guaranteed to match the `Moleclue.name` attribute of the
        molecules in the topology, especially if they are empty strings or not unique.

        See :py:meth:`to_gro <Interchange.to_gro>` and :py:meth:`to_top <Interchange.to_top>` for more details.

        """
        from openff.interchange.interop.gromacs.export._export import GROMACSWriter
        from openff.interchange.smirnoff._gromacs import _convert

        writer = GROMACSWriter(
            system=_convert(self, hydrogen_mass=hydrogen_mass),
            top_file=prefix + ".top",
            gro_file=prefix + ".gro",
        )

        writer.to_top(monolithic=monolithic, _merge_atom_types=_merge_atom_types)
        writer.to_gro(decimal=decimal)

        self.to_mdp(prefix + "_pointenergy.mdp")

    def to_mdp(self, file_path: Path | str):
        """
        Write a GROMACS run configuration ``.MDP`` file for a single-point energy.

        GROMACS considers many of the simulation parameters specified by an
        ``Interchange`` to be run configuration options rather than features of
        the topology. These options are set in the ``.MDP`` file. The written
        ``.MDP`` file includes the appropriate non-bonded configuration for the
        ``Interchange``. The ``nsteps``, ``nstenergy``, and ``continuation``
        configuration values are configured for a single-point energy
        calculation and may be changed as appropriate to perform other
        calculations. See the `GROMACS documentation`_ for details.

        .. _GROMACS documentation: https://manual.gromacs.org/documentation/\
        current/user-guide/mdp-options.html

        Parameters
        ----------
        file_path
            The path to the created GROMACS ``.MDP`` file

        """
        mdconfig = MDConfig.from_interchange(self)
        mdconfig.write_mdp_file(str(file_path))

    def to_top(
        self,
        file_path: Path | str,
        hydrogen_mass: PositiveFloat = 1.007947,
        monolithic: bool = True,
        _merge_atom_types: bool = False,
    ):
        """
        Export this Interchange to a GROMACS topology file.

        Parameters
        ----------
        file_path
            The path to the GROMACS topology file to write.
        hydrogen_mass
            The mass to use for hydrogen atoms if not present in the topology. If non-trivially different
            than the default value, mass will be transferred from neighboring heavy atoms. Note that this is currently
            not applied to any waters and is unsupported when virtual sites are present.
        monolithic
            Whether the topology file should be monolithic (True) or reference individual .itp files (False). Note that
            these individual .itp files rely on ad hoc atom types and cannot be transferred between systems.
        _merge_atom_types
            The flag to define behaviour of GROMACSWriter. If True, then similar atom types will be merged.
            If False, each atom will have its own atom type.

        Notes
        -----
        Molecule names in written files are not guaranteed to match the `Moleclue.name` attribute of the
        molecules in the topology, especially if they are empty strings or not unique.

        """
        from openff.interchange.interop.gromacs.export._export import GROMACSWriter
        from openff.interchange.smirnoff._gromacs import _convert

        GROMACSWriter(
            system=_convert(self, hydrogen_mass=hydrogen_mass),
            top_file=file_path,
            gro_file=tempfile.NamedTemporaryFile(suffix=".gro").file.name,
        ).to_top(
            monolithic=monolithic,
            _merge_atom_types=_merge_atom_types,
        )

    def to_gro(self, file_path: Path | str, decimal: int = 3):
        """
        Export this Interchange object to a GROMACS coordinate file.

        Parameters
        ----------
        file_path
            The path to the GROMACS coordinate file to write.
        decimal
            The number of decimal places to use when writing the GROMACS coordinate file.

        Notes
        -----

        Residue IDs must be positive integers (or string representations thereof).

        Residue IDs greater than 99,999 are reduced modulo 100,000 in line with common GROMACS practice.

        Residue names and IDs from the topology are used, if present, and otherwise are generated internally.

        Behavior when some, but not all, residue names and IDs are present in the topology is undefined.

        If residue IDs are generated internally, they are assigned sequentially to each molecule starting at 1.

        """
        from openff.interchange.interop.gromacs.export._export import GROMACSWriter
        from openff.interchange.smirnoff._gromacs import _convert

        # TODO: Write the coordinates without the full conversion
        GROMACSWriter(
            system=_convert(self),
            top_file=tempfile.NamedTemporaryFile(suffix=".top").file.name,
            gro_file=file_path,
        ).to_gro(decimal=decimal)

    def to_lammps(self, prefix: str, include_type_labels: bool = False):
        """
        Export this ``Interchange`` to LAMMPS data and run input files.

        Parameters
        ----------
        prefix
            The prefix to use for the LAMMPS data and run input files. For
            example, "foo" will produce files named "foo.lmp" and
            "foo_pointenergy.in".
        include_type_labels
            If True, this will include the SMIRKS strings as LAMMPS type labels
            in the LAMMPS data file.
        """
        prefix = str(prefix)
        datafile_path = prefix + ".lmp"
        self.to_lammps_datafile(datafile_path, include_type_labels)
        self.to_lammps_input(
            prefix + "_pointenergy.in",
            datafile_path,
        )

    def to_lammps_datafile(self, file_path: Path | str, include_type_labels: bool = False):
        """Export this Interchange to a LAMMPS data file."""
        from openff.interchange.interop.lammps import to_lammps

        to_lammps(self, file_path, include_type_labels)

    def to_lammps_input(
        self,
        file_path: Path | str,
        data_file: Path | str | None = None,
    ):
        """
        Write a LAMMPS run input file for a single-point energy calculation.

        LAMMPS considers many of the simulation parameters specified by an
        ``Interchange`` to be run configuration options rather than features of
        the force field. These options are set in the run input file.

        Parameters
        ----------
        file_path
            The path to the created LAMMPS run input file
        data_file
            The path to the LAMMPS data file that should be read by the input
            file. If not given, ``file_path`` with the extension ``.lmp`` will
            be used.

        """
        if data_file is None:
            data_file = Path(file_path).with_suffix(".lmp")

        mdconfig = MDConfig.from_interchange(self)
        mdconfig.write_lammps_input(self, str(file_path), data_file=str(data_file))

    def to_openmm_system(
        self,
        combine_nonbonded_forces: bool = True,
        add_constrained_forces: bool = False,
        ewald_tolerance: float = 1e-4,
        hydrogen_mass: PositiveFloat = 1.007947,
    ):
        """
        Export this Interchange to an OpenMM System.

        Parameters
        ----------
        combine_nonbonded_forces
            If True, an attempt will be made to combine all non-bonded interactions into a single
            openmm.NonbondedForce.
            If False, non-bonded interactions will be split across multiple forces.
        add_constrained_forces
            If True, add valence forces that might be overridden by constraints, i.e. call `addBond` or `addAngle`
            on a bond or angle that is fully constrained.
        ewald_tolerance
            The value passed to `NonbondedForce.setEwaldErrorTolerance`
        hydrogen_mass
            The mass to use for hydrogen atoms if not present in the topology. If non-trivially different
            than the default value, mass will be transferred from neighboring heavy atoms. Note that this is currently
            not applied to any waters.

        Returns
        -------
        system
            The OpenMM System object.

        Notes
        -----
        There are some sharp edges and quirks when using this method. Be aware of some documented
        issues in the :doc:`/using/edges` section of the user guide. If you encounter surprising
        behavior that is not documented, please raise an issue.

        """
        from openff.interchange.interop.openmm import (
            to_openmm_system as _to_openmm_system,
        )

        return _to_openmm_system(
            self,
            combine_nonbonded_forces=combine_nonbonded_forces,
            add_constrained_forces=add_constrained_forces,
            ewald_tolerance=ewald_tolerance,
            hydrogen_mass=hydrogen_mass,
        )

    def to_openmm(self, *args, **kwargs):
        return self.to_openmm_system(*args, **kwargs)

    def to_openmm_topology(
        self,
        collate: bool = False,
        ensure_unique_atom_names: str | bool = "residues",
    ):
        """
        Export components of this Interchange to an OpenMM Topology.

        Parameters
        ----------
        collate
            If ``False``, the default, all virtual sites will be added to a
            single residue at the end of the topology. If ``True``, virtual
            sites will be collated with their associated molecule and added to
            the residue of the last atom in the molecule they belong to.

        """
        from openff.interchange.interop.openmm._topology import to_openmm_topology

        return to_openmm_topology(
            self,
            collate=collate,
            ensure_unique_atom_names=ensure_unique_atom_names,
        )

    def to_openmm_simulation(
        self,
        integrator: "openmm.Integrator",
        combine_nonbonded_forces: bool = True,
        add_constrained_forces: bool = False,
        ewald_tolerance: float = 1e-4,
        hydrogen_mass: PositiveFloat = 1.007947,
        additional_forces: Iterable["openmm.Force"] = tuple(),
        **kwargs,
    ) -> "openmm.app.simulation.Simulation":
        """
        Export this Interchange to an OpenMM ``Simulation`` object.

        A :py:class:`Simulation <openmm.app.simulation.Simulation>` encapsulates
        all the information needed for a typical OpenMM simulation into a single
        object with a simple API.

        Positions are set on the ``Simulation`` if present on the
        ``Interchange``.

        Additional forces, such as a barostat, should be added with the
        ``additional_forces`` argument to avoid having to re-initialize
        the ``Context``. Re-initializing the ``Context`` after adding a
        ``Force`` is necessary due to `implementation details`_
        in OpenMM.

        .. _implementation details: https://github.com/openmm/openmm/wiki/Frequently-Asked-Questions#why-does-it-ignore-changes-i-make-to-a-system-or-force

        Parameters
        ----------
        integrator
            The integrator to use for the simulation.
        combine_nonbonded_forces
            If True, an attempt will be made to combine all non-bonded interactions into a single
            openmm.NonbondedForce.
            If False, non-bonded interactions will be split across multiple forces.
        add_constrained_forces
            If True, add valence forces that might be overridden by constraints, i.e. call `addBond` or `addAngle`
            on a bond or angle that is fully constrained.
        ewald_tolerance
            The value passed to `NonbondedForce.setEwaldErrorTolerance`
        hydrogen_mass
            The mass to use for hydrogen atoms if not present in the topology. If non-trivially different
            than the default value, mass will be transferred from neighboring heavy atoms. Note that this is currently
            not applied to any waters.
        additional_forces
            Additional forces to be added to the system, e.g. barostats, that are not
            added by the force field.
        **kwargs
            Further keyword parameters are passed on to
            :py:meth:`Simulation.__init__() <openmm.app.simulation.Simulation.__init__>`

        Returns
        -------
        simulation
            The OpenMM simulation object, possibly with positions set.

        Examples
        --------

        Create an OpenMM simulation with a Langevin integrator and a Monte Carlo barostat:

        >>> import openmm
        >>> import openmm.unit
        >>>
        >>> integrator = openmm.LangevinMiddleIntegrator(
        ...     293.15 * openmm.unit.kelvin,
        ...     1.0 / openmm.unit.picosecond,
        ...     2.0 * openmm.unit.femtosecond,
        ... )
        >>> barostat = openmm.MonteCarloBarostat(
        ...     1.00 * openmm.unit.bar,
        ...     293.15 * openmm.unit.kelvin,
        ...     25,
        ... )
        >>> simulation = interchange.to_openmm_simulation(  # doctest: +SKIP
        ...     integrator=integrator,
        ...     additional_forces=[barostat],
        ... )

        """
        import openmm.app

        system = self.to_openmm_system(
            combine_nonbonded_forces=combine_nonbonded_forces,
            add_constrained_forces=add_constrained_forces,
            ewald_tolerance=ewald_tolerance,
            hydrogen_mass=hydrogen_mass,
        )

        for force in additional_forces:
            system.addForce(force)

        # since we're adding forces before making a context, we don't need to
        # re-initialize context. In a different order, we would need to:
        # https://github.com/openforcefield/openff-interchange/pull/725#discussion_r1210928501

        simulation = openmm.app.Simulation(
            topology=self.to_openmm_topology(),
            system=system,
            integrator=integrator,
            **kwargs,
        )

        # If the system contains virtual sites, the positions must, so no obvious case in which
        # include_virtual_sites could possibly be False
        if self.positions is not None:
            simulation.context.setPositions(
                self.get_positions(include_virtual_sites=True).to_openmm(),
            )

        return simulation

    @requires_package("openmm")
    def to_pdb(self, file_path: Path | str, include_virtual_sites: bool = False):
        """
        Export this Interchange to a .pdb file.

        Note that virtual sites are collated into each molecule, which differs from the default
        behavior of Interchange.to_openmm_topology.
        """
        from openff.interchange.interop.openmm import _to_pdb

        if self.positions is None:
            raise MissingPositionsError(
                "Positions are required to write a `.pdb` file but found None.",
            )

        # TODO: Simply wire `include_virtual_sites` to `to_openmm_{topology|positions}`?
        if include_virtual_sites:
            from openff.interchange.interop._virtual_sites import (
                get_positions_with_virtual_sites,
            )

            openmm_topology = self.to_openmm_topology(
                collate=False,
                ensure_unique_atom_names="residues",
            )
            positions = get_positions_with_virtual_sites(self)

        else:
            openmm_topology = self.topology.to_openmm(
                ensure_unique_atom_names="residues",
            )
            positions = self.positions

        _to_pdb(file_path, openmm_topology, positions.to("angstrom"))

    def to_psf(self, file_path: Path | str):
        """Export this Interchange to a CHARMM-style .psf file."""
        raise UnsupportedExportError

    def to_crd(self, file_path: Path | str):
        """Export this Interchange to a CHARMM-style .crd file."""
        raise UnsupportedExportError

    def to_amber(
        self,
        prefix: str,
    ):
        """
        Export this Interchange object to Amber files.

        Parameters
        ----------
        prefix
            The prefix to use for the Amber parameter/topology, coordinate, and run files, i.e.
            "foo" will produce "foo.top", "foo.gro", and "foo_pointenergy.in".

        Notes
        -----
        The run input file is configured for a single-point energy calculation with sander. It is
        likely portable to pmemd with little or no work.

        """
        self.to_prmtop(f"{prefix}.prmtop")
        self.to_inpcrd(f"{prefix}.inpcrd")

        self.to_sander_input(f"{prefix}_pointenergy.in")

    def to_prmtop(self, file_path: Path | str):
        """Export this Interchange to an Amber .prmtop file."""
        from openff.interchange.interop.amber import to_prmtop

        to_prmtop(self, file_path)

    def to_inpcrd(self, file_path: Path | str):
        """Export this Interchange to an Amber .inpcrd file."""
        from openff.interchange.interop.amber import to_inpcrd

        to_inpcrd(self, file_path)

    def to_sander_input(self, file_path: Path | str):
        """
        Export this ``Interchange`` to a run input file for Amber's SANDER engine.

        Amber considers many of the simulation parameters specified by an
        ``Interchange`` to be run configuration options rather than parameters
        of the topology. These options are set in the SANDER or PMEMD run input
        file. The written SANDER input file includes the appropriate non-bonded
        configuration for the ``Interchange`` which are essential to reproduce
        the desired force field. The file also includes configuration for a
        single-point energy calculation, which should be modified to produce the
        desired simulation.
        """
        mdconfig = MDConfig.from_interchange(self)
        mdconfig.write_sander_input_file(str(file_path))

    @classmethod
    @requires_package("foyer")
    def from_foyer(
        cls,
        force_field: "FoyerForcefield",
        topology: Topology,
        box=None,
        positions=None,
        **kwargs,
    ) -> "Interchange":
        """
        Create an Interchange object from a Foyer force field and an OpenFF topology.

        Examples
        --------
        Generate an Interchange object from a single-molecule (OpenFF) topology and
        the Foyer implementation of OPLS-AA

        .. code-block:: pycon

            >>> from openff.interchange import Interchange
            >>> from openff.toolkit import Molecule, Topology
            >>> from foyer import Forcefield  # doctest: +SKIP
            >>> mol = Molecule.from_smiles("CC")
            >>> mol.generate_conformers(n_conformers=1)
            >>> top = Topology.from_molecules([mol])
            >>> oplsaa = Forcefield(name="oplsaa")  # doctest: +SKIP
            >>> interchange = Interchange.from_foyer(topology=top, force_field=oplsaa)  # doctest: +SKIP
            >>> interchange  # doctest: +SKIP
            Interchange with 8 collections, non-periodic topology with 8 atoms.

        """
        from openff.interchange.foyer._create import _create_interchange

        return _create_interchange(
            force_field=force_field,
            topology=topology,
            box=box,
            positions=positions,
        )

    @classmethod
    @experimental
    def from_gromacs(
        cls,
        topology_file: Path | str,
        gro_file: Path | str,
    ) -> "Interchange":
        """
        Create an Interchange object from GROMACS files.

        Parameters
        ----------
        topology_file
            The path to a GROMACS topology file.
        gro_file
            The path to a GROMACS coordinate file.

        Returns
        -------
        interchange
            An Interchange object representing the contents of the GROMACS files.

        Notes
        -----
        Bond parameters may not correctly be parsed, such as when using SMIRNOFF
        force fields with hydrogen bond constraints.

        """
        from openff.interchange.interop.gromacs._import._import import from_files
        from openff.interchange.interop.gromacs._interchange import to_interchange

        return to_interchange(
            from_files(
                top_file=topology_file,
                gro_file=gro_file,
            ),
        )

    def set_positions_from_gro(
        self,
        gro_file: Path | str,
    ):
        """
        Set the positions of this `Interchange` from a GROMACS coordinate `.gro` file.

        Only the coordinates from the `.gro` file are used. No effort is made to ensure the topologies are compatible
        with each other. This includes, for example, a lack of guarantee that the atom ordering in the `.gro` file
        matches the atom ordering in the `Interchange` object.

        `InvalidPositionsError` is raised if the number of rows in the coordinate array does not match the number of
        atoms in the topology of this `Interchange`.
        """
        from openff.interchange.interop.gromacs._import._import import _read_coordinates

        # should already be in nm, might not be necessary
        coordinates = _read_coordinates(pathlib.Path(gro_file)).to("nanometer")

        if coordinates.shape != (self.topology.n_atoms, 3):
            raise InvalidPositionsError(
                f"Coordinates in {gro_file} do not match the number of atoms in the topology. ",
                f"Parsed coordinates have shape {coordinates.shape} but topology has {self.topology.n_atoms} atoms.",
            )

        self.positions = coordinates

    @classmethod
    def from_openmm(
        cls,
        system: "openmm.System",
        topology: Union["openmm.app.Topology", Topology],
        positions: Quantity | None = None,
        box_vectors: Quantity | None = None,
    ) -> "Interchange":
        """
        Create an Interchange object from OpenMM objects.

        Notes
        -----
        If (topological) bonds in water are missing (physics) parameters, as is often the case with
        rigid water, these parameters will be filled in with values of 1 Angstrom equilibrium bond
        length and a default force constant of 50,000 kcal/mol/A^2, representing an arbitrarily
        stiff harmonic bond, and angle parameters of 104.5 degrees and 1.0 kcal/mol/rad^2,
        representing an arbitrarily harmonic angle. It is expected that these values will be
        overwritten by runtime MD options.

        See more limitations and sharp edges in the user guide: https://docs.openforcefield.org/projects/interchange/en/latest/using/edges.html#quirks-of-from-openmm

        Parameters
        ----------
        system
            The OpenMM system.
        topology
            The OpenMM topology.
        positions
            The positions of particles in this system and/or topology.
        box_vectors
            The vectors of the simulation box associated with this system and/or topology.

        Returns
        -------
        interchange
            An Interchange object representing the contents of the OpenMM objects.

        Notes
        -----
        An `openmm.CMMotionRemover` force, if present, is ignored.

        """
        from openff.interchange.interop.openmm._import._import import from_openmm

        return from_openmm(
            topology=topology,
            system=system,
            positions=positions,
            box_vectors=box_vectors,
        )

    def _get_parameters(self, handler_name: str, atom_indices: tuple[int]) -> dict:
        """
        Get parameter values of a specific potential.

        Here, parameters are expected to be uniquely dfined by the name of
        its associated handler and a tuple of atom indices.

        Note: This method only checks for equality of atom indices and will likely fail on complex cases
        involved layered parameters with multiple topology keys sharing identical atom indices.
        """
        for handler in self.collections:
            if handler == handler_name:
                return self[handler_name]._get_parameters(atom_indices=atom_indices)
        raise MissingParameterHandlerError(
            f"Could not find parameter handler of name {handler_name}",
        )

    @overload
    def __getitem__(self, item: Literal["Bonds"]) -> "BondCollection": ...

    @overload
    def __getitem__(
        self,
        item: Literal["Constraints"],
    ) -> "SMIRNOFFConstraintCollection": ...

    @overload
    def __getitem__(self, item: Literal["Angles"]) -> "AngleCollection": ...

    @overload
    def __getitem__(
        self,
        item: Literal["vdW"],
    ) -> "vdWCollection": ...

    @overload
    def __getitem__(
        self,
        item: Literal["ProperTorsions"],
    ) -> "ProperTorsionCollection": ...

    @overload
    def __getitem__(
        self,
        item: Literal["ImproperTorsions"],
    ) -> "ImproperTorsionCollection": ...

    @overload
    def __getitem__(
        self,
        item: Literal["VirtualSites"],
    ) -> "SMIRNOFFVirtualSiteCollection": ...

    @overload
    def __getitem__(
        self,
        item: Literal["Electrostatics"],
    ) -> "ElectrostaticsCollection": ...

    @overload
    def __getitem__(
        self,
        item: Literal["GBSA"],
    ) -> "SMIRNOFFGBSACollection": ...

    @overload
    def __getitem__(
        self,
        item: str,
    ) -> "Collection": ...

    def __getitem__(self, item: str):
        """Syntax sugar for looking up collections or other components."""
        if type(item) is not str:
            raise LookupError(
                f"Only str arguments can be currently be used for lookups.\nFound item {item} of type {type(item)}",
            )
        if item == "positions":
            return self.positions
        elif item in {"box", "box_vectors"}:
            return self.box
        elif item in self.collections:
            return self.collections[item]
        else:
            raise LookupError(
                f"Could not find component {item}. This object has the following "
                f"collections registered:\n\t{[*self.collections.keys()]}",
            )

    def __add__(self, other: "Interchange") -> "Interchange":
        """Combine two Interchange objects."""
        warnings.warn(
            "The `+` operator is deprecated. Use `Interchange.combine` instead.",
            InterchangeDeprecationWarning,
        )

        return self.combine(other)

    def combine(self, other: "Interchange") -> "Interchange":
        """
        Combine two Interchange objects.

        Parameters
        ----------
        other
            The other Interchange object to combine with this one.

        Returns
        -------
        An Interchange object reprensenting a combination of the inputs. The topology is modified by
        appending the "other" topology to the "self" topology. No atoms are reorderded.

        Notes
        -----
        Interchange object combination is complex and may produce strange
        results outside of use cases it has been tested in. Use with caution and
        thoroughly validate results!

        The non-bonded methods (i.e. non-bonded cutoffs, use of a switching
        function, etc.) must be compatible between the two inputs. Otherwise, an error is raised.

        Some topology and potential keys used in parameter bookkeeping may be
        modified to work around edge cases in which some parameter values may be
        mangled.

        """
        from openff.interchange.operations._combine import _combine

        return _combine(self, other)

    def __repr__(self) -> str:
        periodic = self.box is not None
        n_atoms = self.topology.n_atoms
        return (
            f"Interchange with {len(self.collections)} collections, "
            f"{'' if periodic else 'non-'}periodic topology with {n_atoms} atoms."
        )



import pandas
import pytest
from openff.toolkit import ForceField, Molecule, Quantity
from openff.utilities.testing import skip_if_missing

from openff.interchange.components.mdconfig import get_intermol_defaults
from openff.interchange.drivers import get_summary_data
from openff.interchange.exceptions import UnsupportedExportError


def run(smiles: str, constrained: bool) -> pandas.DataFrame:
    molecule = Molecule.from_smiles(smiles)
    molecule.generate_conformers(n_conformers=1)
    topology = molecule.to_topology()
    topology.box_vectors = Quantity([4, 4, 4], "nanometer")

    if constrained:
        force_field = ForceField("openff-2.0.0.offxml")
    else:
        force_field = ForceField("openff_unconstrained-2.0.0.offxml")

    interchange = force_field.create_interchange(topology)
    get_intermol_defaults(periodic=True).apply(interchange)
    return get_summary_data(interchange)


def compare(smiles: str) -> float:
    return run(smiles, True).std()["Bond"] - run(smiles, False).std()["Bond"]


@skip_if_missing("openmm")
class TestToOpenMM:
    def test_cmm_remover_included(self, sage, basic_top):
        import openmm

        system = sage.create_interchange(basic_top).to_openmm_system()

        assert isinstance(
            system.getForce(system.getNumForces() - 1),
            openmm.CMMotionRemover,
        )

    def test_combine_nonbonded_forces_vdw_14(self, sage_unconstrained):
        molecule = Molecule.from_mapped_smiles("[C:2](#[C:3][Br:4])[Br:1]")
        molecule.generate_conformers(n_conformers=1)
        topology = molecule.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        interchange = sage_unconstrained.create_interchange(topology)

        split_system = interchange.to_openmm(combine_nonbonded_forces=False)

        vdw_14_force = next(force for force in split_system.getForces() if force.getName() == "vdW 1-4 force")
        vdw_force = next(force for force in split_system.getForces() if force.getName() == "vdW force")

        assert vdw_force.getParticleParameters(0) == vdw_force.getParticleParameters(3)

        _, _, (sigma, epsilon) = vdw_14_force.getBondParameters(0)

        # epsilon of 1-4 interaction matches either atom's sigma
        assert sigma == vdw_force.getParticleParameters(0)[0]

        # epsilon of 1-4 interaction matches either atom's sigma
        scale = sage_unconstrained["vdW"].scale14
        assert epsilon == vdw_force.getParticleParameters(0)[1] * scale

    def test_virtual_sites_no_vdw_error(self, tip4p, water):
        tip4p.deregister_parameter_handler("vdW")
        with pytest.raises(
            UnsupportedExportError,
            match="Virtual sites with no vdW handler",
        ):
            tip4p.create_interchange(water.to_topology()).to_openmm_system()

    def test_particles_exist_without_nonbonded_force(self, sage_unconstrained):
        import openmm

        valence_ff = ForceField()

        valence_ff.register_parameter_handler(sage_unconstrained["Bonds"])
        valence_ff.register_parameter_handler(sage_unconstrained["Angles"])

        valence_ff.register_parameter_handler(sage_unconstrained["ProperTorsions"])
        valence_ff.register_parameter_handler(sage_unconstrained["ImproperTorsions"])

        interchange = valence_ff.create_interchange(Molecule.from_smiles("CCO").to_topology())

        assert "vdW" not in interchange.collections
        assert "Electrostatics" not in interchange.collections

        openmm_system = interchange.to_openmm()

        # do we want to check the PDB file / OpenMM topology exported fro this state, too?

        assert openmm_system.getNumParticles() == 9

        for force in openmm_system.getForces():
            assert not isinstance(force, openmm.NonbondedForce | openmm.CustomNonbondedForce)
            assert isinstance(
                force,
                openmm.HarmonicBondForce
                | openmm.HarmonicAngleForce
                | openmm.PeriodicTorsionForce
                | openmm.CMMotionRemover,
            )

    def test_particles_exist_with_no_forces(self):
        import openmm

        empty_ff = ForceField()

        assert len(empty_ff.registered_parameter_handlers) == 0

        interchange = empty_ff.create_interchange(Molecule.from_smiles("CCO").to_topology())

        # Interchange creates an empty ConstraintsCollection by default, even with no
        # constraints handler or other handlers that would populate it
        assert len(interchange.collections) == 1

        openmm_system = interchange.to_openmm()

        assert openmm_system.getNumParticles() == 9

        # CMMotionRemover is added by default, even with no forces
        assert openmm_system.getNumForces() == 1

        assert type(openmm_system.getForce(0)) is openmm.CMMotionRemover



import pytest

from openff.interchange._experimental import experimental
from openff.interchange.exceptions import ExperimentalFeatureException


def test_default():
    @experimental
    def f():
        """Docstring of f."""

    with pytest.raises(ExperimentalFeatureException):
        f()


def test_experimental_opted_in(monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    @experimental
    def g():
        """Docstring of g."""

    g()


def test_experimental_opted_in_bad_value(monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "True")

    @experimental
    def h():
        """Docstring of h."""

    with pytest.raises(ExperimentalFeatureException):
        h()



import copy
import functools
import logging
import warnings
from collections.abc import Iterable
from typing import Literal, Self

import numpy
from openff.toolkit import GLOBAL_TOOLKIT_REGISTRY, Molecule, NAGLToolkitWrapper, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff.parameters import (
    ChargeIncrementModelHandler,
    ElectrostaticsHandler,
    LibraryChargeHandler,
    NAGLChargesHandler,
    ParameterHandler,
    ToolkitAM1BCCHandler,
    vdWHandler,
)
from openff.toolkit.utils.exceptions import MissingPackageError, SMIRNOFFSpecError
from packaging.version import Version
from pydantic import Field, PrivateAttr, computed_field

from openff.interchange._annotations import _ElementaryChargeQuantity
from openff.interchange.common._nonbonded import (
    ElectrostaticsCollection,
    _NonbondedCollection,
    vdWCollection,
)
from openff.interchange.components.potentials import Potential
from openff.interchange.exceptions import (
    InvalidParameterHandlerError,
    MissingPartialChargesError,
    NonIntegralMoleculeChargeError,
    SMIRNOFFVersionNotSupportedError,
)
from openff.interchange.models import (
    ChargeIncrementTopologyKey,
    ChargeModelTopologyKey,
    LibraryChargeTopologyKey,
    PotentialKey,
    SingleAtomChargeTopologyKey,
    TopologyKey,
    VirtualSiteKey,
)
from openff.interchange.smirnoff._base import SMIRNOFFCollection
from openff.interchange.warnings import ForceFieldModificationWarning

logger = logging.getLogger(__name__)

type ElectrostaticsHandlerType = (
    ElectrostaticsHandler
    | ToolkitAM1BCCHandler
    | ChargeIncrementModelHandler
    | LibraryChargeHandler
    | NAGLChargesHandler
)


_ZERO_CHARGE = Quantity(0.0, "elementary_charge")


@functools.lru_cache(None)
def _add_charges(
    charge1: float,
    charge2: float,
) -> "Quantity":
    """Add two charges together."""
    return Quantity(charge1 + charge2, "elementary_charge")


def _upconvert_vdw_handler(vdw_handler: vdWHandler):
    """Given a vdW with version 0.3 or 0.4, up-convert to 0.4 or short-circuit if already 0.4."""
    if vdw_handler.version >= Version("0.4"):
        return

    elif vdw_handler.version == Version("0.3"):
        warnings.warn(
            "Automatically up-converting vdWHandler from version 0.3 to 0.4. Consider manually upgrading "
            "this vdW (or <vdW> section in an OFFXML file) to 0.4 or newer. For more details, "
            "see https://openforcefield.github.io/standards/standards/smirnoff/#vdw.",
            ForceFieldModificationWarning,
        )

        if vdw_handler.method != "cutoff":
            raise NotImplementedError(
                "Up-converting vdWHandler with method != 'cutoff' not yet supported.",
            )

        vdw_handler.version = Version("0.4")
        vdw_handler.periodic_method = "cutoff"
        vdw_handler.nonbonded_method = "no-cutoff"

        delattr(vdw_handler, "method")


def _downconvert_vdw_handler(vdw_handler: vdWHandler):
    """Given a vdWHandler with version 0.4, down-convert to 0.3."""
    if vdw_handler.version == Version("0.3"):
        return

    elif vdw_handler.version == Version("0.4"):
        warnings.warn(
            "Automatically down-converting vdWHandler from version 0.4 to 0.3. In the future, this "
            "down-conversion will not happen and version 0.3 will not be supported.",
            ForceFieldModificationWarning,
        )

        if (vdw_handler.periodic_method != "cutoff") or (vdw_handler.nonperiodic_method != "no-cutoff"):
            raise NotImplementedError("Down-converting vdWHandler failed.")

        vdw_handler.version = Version("0.3")
        vdw_handler.method = "cutoff"

        try:
            delattr(vdw_handler, "periodic_method")
            delattr(vdw_handler, "nonperiodic_method")
        except AttributeError:
            # If they cannot be deleted, just leave them hanging around. Not ideal!
            # https://github.com/openforcefield/openff-toolkit/issues/1680
            pass


def library_charge_from_molecule(
    molecule: "Molecule",
) -> LibraryChargeHandler.LibraryChargeType:
    """Given an OpenFF Molecule with charges, generate a corresponding LibraryChargeType."""
    if molecule.partial_charges is None:
        raise MissingPartialChargesError

    smirks = molecule.to_smiles(mapped=True)
    charges = molecule.partial_charges

    library_charge_type = LibraryChargeHandler.LibraryChargeType(
        smirks=smirks,
        charge=charges,
    )

    return library_charge_type


class _SMIRNOFFNonbondedCollection(SMIRNOFFCollection, _NonbondedCollection):
    """Base class for handlers storing non-bonded potentials produced by SMIRNOFF force fields."""


class SMIRNOFFvdWCollection(vdWCollection, SMIRNOFFCollection):
    """Handler storing vdW potentials as produced by a SMIRNOFF force field."""

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [vdWHandler]

    @classmethod
    def supported_parameters(cls) -> Iterable[str]:
        """Return a list of supported parameter attributes."""
        return ["smirks", "id", "sigma", "epsilon", "rmin_half"]

    @classmethod
    def default_parameter_values(cls) -> Iterable[float]:
        """Per-particle parameter values passed to Force.addParticle()."""
        return [1.0, 0.0]

    @classmethod
    def potential_parameters(cls) -> Iterable[str]:
        """Return a list of names of parameters included in each potential in this colletion."""
        return ["sigma", "epsilon"]

    def store_potentials(self, parameter_handler: vdWHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [TopologyKey, PotentialKey].

        """
        self.periodic_method = parameter_handler.periodic_method.lower().replace(
            "ewald3d",
            "pme",
        )
        self.nonperiodic_method = parameter_handler.nonperiodic_method.lower()
        self.cutoff = parameter_handler.cutoff

        for potential_key in self.key_map.values():
            smirks = potential_key.id
            force_field_parameters = parameter_handler.parameters[smirks]

            potential = Potential(
                parameters={
                    parameter: getattr(force_field_parameters, parameter) for parameter in self.potential_parameters()
                },
            )

            self.potentials[potential_key] = potential

    @classmethod
    def create(
        cls,
        parameter_handler: vdWHandler,
        topology: Topology,
    ) -> Self:
        """
        Create a SMIRNOFFvdWCollection from toolkit data.

        """
        if isinstance(parameter_handler, list):
            parameter_handlers = parameter_handler
        else:
            parameter_handlers = [parameter_handler]

        for parameter_handler in parameter_handlers:
            if type(parameter_handler) not in cls.allowed_parameter_handlers():
                raise InvalidParameterHandlerError(
                    f"Found parameter handler type {type(parameter_handler)}, which is not "
                    f"supported by potential type {type(cls)}",
                )

        handler = cls(
            scale_12=parameter_handler.scale12,
            scale_13=parameter_handler.scale13,
            scale_14=parameter_handler.scale14,
            scale_15=parameter_handler.scale15,
            cutoff=parameter_handler.cutoff,
            mixing_rule=parameter_handler.combining_rules.lower(),
            periodic_method=parameter_handler.periodic_method.lower().replace(
                "ewald3d",
                "pme",
            ),
            nonperiodic_method=parameter_handler.nonperiodic_method.lower(),
            switch_width=parameter_handler.switch_width,
        )
        handler.store_matches(parameter_handler=parameter_handler, topology=topology)
        handler.store_potentials(parameter_handler=parameter_handler)

        return handler

    @classmethod
    def parameter_handler_precedence(cls) -> list[str]:
        """
        Return the order in which parameter handlers take precedence when computing charges.
        """
        return ["vdw", "VirtualSites"]


class SMIRNOFFElectrostaticsCollection(ElectrostaticsCollection, SMIRNOFFCollection):
    """
    A handler which stores any electrostatic parameters applied to a topology.

    This handler is responsible for grouping together

    * global settings for the electrostatic interactions such as the cutoff distance
      and the intramolecular scale factors.
    * partial charges which have been assigned by a ``ToolkitAM1BCC``,
      ``LibraryCharges``, ``NAGLCharges``, or a ``ChargeIncrementModel`` parameter
      handler.
    * charge corrections applied by a ``ChargeIncrementHandler``

    rather than having each in their own handler.
    """

    periodic_potential: Literal[
        "Ewald3D-ConductingBoundary",
        "cutoff",
        "no-cutoff",
        "reaction-field",
    ] = Field("Ewald3D-ConductingBoundary")
    nonperiodic_potential: Literal[
        "Coulomb",
        "cutoff",
        "no-cutoff",
        "reaction-field",
    ] = Field("Coulomb")
    exception_potential: Literal["Coulomb"] = Field("Coulomb")

    _charges: dict[
        TopologyKey | LibraryChargeTopologyKey | VirtualSiteKey,
        _ElementaryChargeQuantity,
    ] = PrivateAttr(dict())
    _charges_cached: bool = PrivateAttr(default=False)

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [
            LibraryChargeHandler,
            NAGLChargesHandler,
            ChargeIncrementModelHandler,
            ToolkitAM1BCCHandler,
            ElectrostaticsHandler,
        ]

    @classmethod
    def supported_parameters(cls):
        """Return a list of supported parameter attribute names."""

    @property
    def _charges_without_virtual_sites(
        self,
    ) -> dict[TopologyKey | LibraryChargeTopologyKey, _ElementaryChargeQuantity]:
        """Get the total partial charge on each atom, excluding virtual sites."""
        return self._get_charges(include_virtual_sites=False)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def charges(
        self,
    ) -> dict[TopologyKey | LibraryChargeTopologyKey | VirtualSiteKey, _ElementaryChargeQuantity]:
        """
        Get the total partial charge on each atom, including virtual sites, as a rich dict.

        For a simpler array-like representation, see `get_charge_array`.
        """
        if len(self._charges) == 0 or self._charges_cached is False:
            # TODO: Clearing this dict **should not be necessary** but in some cases in appears
            #       that this class attribute persists in some cases, possibly only on a single
            #       thread. Ideas to try include
            #           * Drop @computed_field ?
            #           * Don't handle caching ourselves and let Pydantic do it (i.e. have this
            #               function simply retrun ._get_charges(...)
            #
            #       Hopefully this isn't a major issue - caches for large systems should still
            #       only be build once
            #
            #       Some more context: https://github.com/openforcefield/openff-interchange/issues/842#issuecomment-2394211357
            self._charges.clear()

            self._charges.update(self._get_charges(include_virtual_sites=True))
            self._charges_cached = True

        return self._charges

    def _get_charges(
        self,
        include_virtual_sites: bool = True,
    ) -> dict[TopologyKey | LibraryChargeTopologyKey | VirtualSiteKey, _ElementaryChargeQuantity]:
        """Get the total partial charge on each atom or particle."""
        # Keyed by index for atoms and by VirtualSiteKey for virtual sites.
        # work in unitless (float, implicit elementary_charge) values until returning
        charges: dict[VirtualSiteKey | int, _ElementaryChargeQuantity] = dict()

        for topology_key, potential_key in self.key_map.items():
            potential = self.potentials[potential_key]

            for parameter_key, parameter_value in potential.parameters.items():
                if parameter_key == "charge_increments":
                    if type(topology_key) is not VirtualSiteKey:
                        raise RuntimeError(
                            '"charge_increments" (plural) understood to only be applied by'
                            "virtual sites, not by a `ChargeIncrementModelHandler`.",
                        )

                    # assumes virtual sites can only have charges determined in one step
                    charges[topology_key] = -1.0 * numpy.sum(parameter_value)

                    # Apply increments to "orientation" atoms
                    for i, increment in enumerate(parameter_value):
                        orientation_atom_index = topology_key.orientation_atom_indices[i]

                        charges[orientation_atom_index] = _add_charges(
                            charges.get(orientation_atom_index, _ZERO_CHARGE).m,
                            increment.m,
                        )

                elif parameter_key == "charge":
                    assert len(topology_key.atom_indices) == 1

                    atom_index = topology_key.atom_indices[0]

                    if potential_key.associated_handler in (
                        "LibraryCharges",
                        "NAGLChargesHandler",
                        "ToolkitAM1BCCHandler",
                        "molecules_with_preset_charges",
                        "ExternalSource",
                    ):
                        charges[atom_index] = _add_charges(
                            charges.get(atom_index, _ZERO_CHARGE).m,
                            parameter_value.m,
                        )

                    elif potential_key.associated_handler in (  # type: ignore[operator]
                        "ChargeIncrementModelHandler"
                    ):
                        # the "charge" and "charge_increment" keys may not appear in that order, so
                        # we "add" the charge whether or not the increment was already applied.
                        # There should be a better way to do this.

                        charges[atom_index] = _add_charges(
                            charges.get(atom_index, _ZERO_CHARGE).m,
                            parameter_value.m,
                        )

                    else:
                        raise RuntimeError(
                            f"Unexepected associated handler {potential_key.associated_handler} found.",
                        )

                elif parameter_key == "charge_increment":
                    assert type(topology_key) is ChargeIncrementTopologyKey
                    assert len(topology_key.atom_indices) == 1

                    atom_index = topology_key.atom_indices[0]

                    charges[atom_index] = _add_charges(
                        charges.get(atom_index, _ZERO_CHARGE).m,
                        parameter_value.m,
                    )

                    logger.debug(
                        "Charge section ChargeIncrementModel, applying charge increment from atom "
                        f"{topology_key.this_atom_index} to atoms {topology_key.other_atom_indices}",
                    )

                else:
                    raise NotImplementedError()

        returned_charges: dict[
            TopologyKey | LibraryChargeTopologyKey | VirtualSiteKey,
            Quantity,
        ] = dict()

        for index, charge in charges.items():
            if type(index) is int:
                # Note this uses the topology index, even if the particle index differs in an export,
                # i.e. if a molecule earlier in the topology has (collated) virtual sites as is
                # common with GROMACS files
                returned_charges[TopologyKey(atom_indices=(index,))] = charge
            elif type(index) is VirtualSiteKey:
                if include_virtual_sites:
                    returned_charges[index] = charge

        return {key: Quantity(val, "elementary_charge") for key, val in returned_charges.items()}

    @classmethod
    def parameter_handler_precedence(cls) -> list[str]:
        """
        Return the order in which parameter handlers take precedence when computing charges.
        """
        return ["LibraryCharges", "NAGLCharges", "ChargeIncrementModel", "ToolkitAM1BCC"]

    @classmethod
    def create(
        cls,
        parameter_handler: ParameterHandler | list[ParameterHandler],
        topology: Topology,
        molecules_with_preset_charges=None,
        allow_nonintegral_charges: bool = False,
    ) -> Self:
        """
        Create a SMIRNOFFElectrostaticsCollection from toolkit data.

        """
        if isinstance(parameter_handler, list):
            parameter_handlers = parameter_handler
        else:
            parameter_handlers = [parameter_handler]

        for handler in parameter_handlers:
            if isinstance(handler, ElectrostaticsHandler):
                if Version(str(handler.version)) < Version("0.4"):
                    raise SMIRNOFFVersionNotSupportedError(
                        "Electrostatics section must be up-converted to 0.4 or newer. Found version "
                        f"{handler.version}.",
                    )

        toolkit_handler_with_metadata = next(p for p in parameter_handlers if type(p) is ElectrostaticsHandler)

        handler = cls(
            type=toolkit_handler_with_metadata.TAGNAME,
            scale_12=toolkit_handler_with_metadata.scale12,
            scale_13=toolkit_handler_with_metadata.scale13,
            scale_14=toolkit_handler_with_metadata.scale14,
            scale_15=toolkit_handler_with_metadata.scale15,
            cutoff=toolkit_handler_with_metadata.cutoff,
            periodic_potential=toolkit_handler_with_metadata.periodic_potential,
            nonperiodic_potential=toolkit_handler_with_metadata.nonperiodic_potential,
            exception_potential=toolkit_handler_with_metadata.exception_potential,
        )

        handler.store_matches(
            parameter_handlers,
            topology,
            molecules_with_preset_charges=molecules_with_preset_charges,
            allow_nonintegral_charges=allow_nonintegral_charges,
        )

        handler._charges = dict()

        return handler

    @classmethod
    @functools.lru_cache(None)
    def _compute_partial_charges(
        cls,
        molecule: Molecule,
        mapped_smiles: str,
        method: str,
        exception_types_to_raise: Iterable[Exception],
        additional_args: tuple[tuple[str, str]],
    ) -> Quantity:
        """Call out to the toolkit's toolkit wrappers to generate partial charges."""
        additional_args_as_dict: dict[str, str] = dict(additional_args)
        molecule = copy.deepcopy(molecule)
        GLOBAL_TOOLKIT_REGISTRY.call(
            "assign_partial_charges",
            molecule=molecule,
            partial_charge_method=method,
            raise_exception_types=exception_types_to_raise,
            **additional_args_as_dict,
        )

        return molecule.partial_charges

    @classmethod
    def _library_charge_to_potentials(
        cls,
        atom_indices: tuple[int, ...],
        parameter: LibraryChargeHandler.LibraryChargeType,
    ) -> tuple[
        dict[LibraryChargeTopologyKey, PotentialKey],
        dict[PotentialKey, Potential],
    ]:
        """
        Map a matched library charge parameter to a set of potentials.
        """
        matches = {}
        potentials = {}

        for i, (atom_index, charge) in enumerate(zip(atom_indices, parameter.charge)):
            topology_key = LibraryChargeTopologyKey(this_atom_index=atom_index)
            potential_key = PotentialKey(
                id=parameter.smirks,
                mult=i,
                associated_handler="LibraryCharges",
            )
            potential = Potential(parameters={"charge": charge})

            matches[topology_key] = potential_key
            potentials[potential_key] = potential

        return matches, potentials

    @classmethod
    def _charge_increment_to_potentials(
        cls,
        atom_indices: tuple[int, ...],
        parameter: ChargeIncrementModelHandler.ChargeIncrementType,
    ) -> tuple[
        dict[ChargeIncrementTopologyKey, PotentialKey],
        dict[PotentialKey, Potential],
    ]:
        """
        Map a matched charge increment parameter to a set of potentials.
        """
        matches = {}
        potentials = {}

        for i, atom_index in enumerate(atom_indices):
            other_atom_indices = tuple(val for val in atom_indices if val is not atom_index)
            topology_key = ChargeIncrementTopologyKey(
                this_atom_index=atom_index,
                other_atom_indices=other_atom_indices,
            )
            # TopologyKey(atom_indices=(atom_index,), mult=other_index)
            potential_key = PotentialKey(
                id=parameter.smirks,
                mult=i,
                associated_handler="ChargeIncrementModel",
            )

            # TODO: Handle the cases where n - 1 charge increments have been defined,
            #       maybe by implementing this in the TK?
            charge_increment = getattr(parameter, f"charge_increment{i + 1}")

            potential = Potential(parameters={"charge_increment": charge_increment})

            matches[topology_key] = potential_key
            potentials[potential_key] = potential

        return matches, potentials

    @classmethod
    def _find_slot_matches(
        cls,
        parameter_handler: LibraryChargeHandler | ChargeIncrementModelHandler,
        unique_molecule: Molecule,
    ) -> tuple[dict[TopologyKey, PotentialKey], dict[PotentialKey, Potential]]:
        """
        Construct a slot and potential map for a slot based parameter handler.
        """
        # Ideally this would be made redundant by OpenFF TK #971
        unique_parameter_matches = {
            tuple(sorted(key)): (key, val)
            for key, val in parameter_handler.find_matches(
                unique_molecule.to_topology(),
            ).items()
        }

        parameter_matches = {key: val for key, val in unique_parameter_matches.values()}
        if type(parameter_handler) is ChargeIncrementModelHandler:
            for atom_indices, val in parameter_matches.items():
                charge_increments = val.parameter_type.charge_increment

                if len(atom_indices) - len(charge_increments) == 0:
                    pass
                elif len(atom_indices) - len(charge_increments) == 1:
                    # If we've been provided with one less charge increment value than tagged atoms, assume the last
                    # tagged atom offsets the charge of the others to make the chargeincrement net-neutral
                    charge_increment_sum = Quantity(0.0, "elementary_charge")

                    for ci in charge_increments:
                        charge_increment_sum += ci
                    charge_increments.append(-charge_increment_sum)

                else:
                    raise SMIRNOFFSpecError(
                        f"Trying to apply chargeincrements {val.parameter_type} "
                        f"to tagged atoms {atom_indices}, but the number of chargeincrements "
                        "must be either the same as- or one less than the number of tagged atoms."
                        f"found {len(atom_indices)} number of tagged atoms and "
                        f"{len(val.parameter_type.charge_increment)} number of charge increments",
                    )

        matches, potentials = {}, {}

        for key, val in parameter_matches.items():
            parameter = val.parameter_type

            if isinstance(parameter_handler, LibraryChargeHandler):
                (
                    parameter_matches,
                    parameter_potentials,
                ) = cls._library_charge_to_potentials(key, parameter)

            elif isinstance(parameter_handler, ChargeIncrementModelHandler):
                (
                    parameter_matches,
                    parameter_potentials,
                ) = cls._charge_increment_to_potentials(key, parameter)

            else:
                raise NotImplementedError()

            for topology_key, potential_key in parameter_matches.items():
                # This may silently overwrite an identical key generated from a previous match, but that is
                # the toolkit behavior (see test_assign_charges_to_molecule_in_parts_using_multiple_library_charges).
                # If there is a need to track the topology keys that are ignored, this can be changed.
                matches[topology_key] = potential_key

            potentials.update(parameter_potentials)

        return matches, potentials  # type: ignore[return-value]

    @classmethod
    def _find_charge_model_matches(
        cls,
        parameter_handler: ToolkitAM1BCCHandler | ChargeIncrementModelHandler | NAGLChargesHandler,
        unique_molecule: Molecule,
    ) -> tuple[
        str,
        dict[SingleAtomChargeTopologyKey, PotentialKey],
        dict[PotentialKey, Potential],
    ]:
        """Construct a slot and potential map for a charge model based parameter handler."""

        unique_molecule = copy.deepcopy(unique_molecule)
        reference_smiles = unique_molecule.to_smiles(
            isomeric=True,
            explicit_hydrogens=True,
            mapped=True,
        )

        handler_name = parameter_handler.__class__.__name__

        if handler_name == "NAGLChargesHandler":
            from openff.nagl_models._dynamic_fetch import HashComparisonFailedException, UnableToParseDOIException

            partial_charge_method = parameter_handler.model_file
            if NAGLToolkitWrapper not in {type(tk) for tk in GLOBAL_TOOLKIT_REGISTRY.registered_toolkits}:
                raise MissingPackageError(
                    "The force field has a NAGLCharges section, but the NAGL software isn't "
                    "present in GLOBAL_TOOLKIT_REGISTRY",
                )
            exception_types_to_raise: tuple = (
                FileNotFoundError,
                MissingPackageError,
                HashComparisonFailedException,
                UnableToParseDOIException,
            )
            additional_args: tuple = (
                ("doi", parameter_handler.digital_object_identifier),
                ("file_hash", parameter_handler.model_file_hash),
            )
        elif handler_name == "ChargeIncrementModelHandler":
            partial_charge_method = parameter_handler.partial_charge_method
            exception_types_to_raise = tuple()
            additional_args = tuple()
        elif handler_name == "ToolkitAM1BCCHandler":
            # The implementation of _toolkit_registry_manager should result in this `GLOBAL_TOOLKIT_REGISTRY`
            # including only what it is passed, even if it's not what one would expect at import time
            if "OpenEye" in GLOBAL_TOOLKIT_REGISTRY.__repr__():
                partial_charge_method = "am1bccelf10"
            else:
                partial_charge_method = "am1bcc"
            exception_types_to_raise = tuple()
            additional_args = tuple()
        else:
            raise InvalidParameterHandlerError(
                f"Encountered unknown handler of type {type(parameter_handler)} where only "
                "ToolkitAM1BCCHandler, NAGLChargesHandler, or ChargeIncrementModelHandler are expected.",
            )
        partial_charges = cls._compute_partial_charges(
            unique_molecule,
            unique_molecule.to_smiles(
                isomeric=True,
                explicit_hydrogens=True,
                mapped=True,
            ),
            method=partial_charge_method,
            exception_types_to_raise=exception_types_to_raise,
            additional_args=additional_args,
        )
        matches = {}
        potentials = {}

        for atom_index, partial_charge in enumerate(partial_charges):
            # These arguments make this object specific to this atom (by index) in this molecule ONLY
            # (assuming an isomeric, mapped, explicit hydrogen SMILES is unique, which seems true).
            potential_key = PotentialKey(
                id=reference_smiles,
                mult=atom_index,
                associated_handler=handler_name,
            )
            potentials[potential_key] = Potential(parameters={"charge": partial_charge})

            matches[
                SingleAtomChargeTopologyKey(
                    this_atom_index=atom_index,
                    extras={
                        "handler": handler_name,
                        "partial_charge_method": partial_charge_method,
                    },
                )
            ] = potential_key

        return partial_charge_method, matches, potentials

    @classmethod
    def _find_reference_matches(
        cls,
        parameter_handlers: dict[str, ElectrostaticsHandlerType],
        unique_molecule: Molecule,
    ) -> tuple[dict[TopologyKey, PotentialKey], dict[PotentialKey, Potential]]:
        """
        Construct a slot and potential map for a particular reference molecule and set of parameter handlers.
        """
        matches: dict[TopologyKey, PotentialKey] = dict()
        potentials: dict[PotentialKey, Potential] = dict()

        expected_matches = {i for i in range(unique_molecule.n_atoms)}

        for handler_type in cls.parameter_handler_precedence():
            if handler_type not in parameter_handlers:
                continue

            parameter_handler = parameter_handlers[handler_type]

            slot_matches, am1_matches = None, None
            slot_potentials: dict = {}
            am1_potentials: dict = {}

            if handler_type in ["LibraryCharges", "ChargeIncrementModel"]:
                slot_matches, slot_potentials = cls._find_slot_matches(
                    parameter_handler,
                    unique_molecule,
                )

            if handler_type in ["ToolkitAM1BCC", "ChargeIncrementModel", "NAGLCharges"]:
                (
                    partial_charge_method,
                    am1_matches,
                    am1_potentials,
                ) = cls._find_charge_model_matches(
                    parameter_handler,
                    unique_molecule,
                )

            if slot_matches is None and am1_matches is None:
                raise NotImplementedError()

            elif slot_matches is not None and am1_matches is not None:
                am1_matches = {
                    ChargeModelTopologyKey(  # type: ignore[misc]
                        this_atom_index=topology_key.atom_indices[0],
                        partial_charge_method=partial_charge_method,
                    ): potential_key
                    for topology_key, potential_key in am1_matches.items()
                }

                matched_atom_indices = {index for key in slot_matches for index in key.atom_indices}
                matched_atom_indices.update(
                    {index for key in am1_matches for index in key.atom_indices},
                )

            elif slot_matches is not None:
                matched_atom_indices = {index for key in slot_matches for index in key.atom_indices}
            else:
                matched_atom_indices = {
                    index
                    for key in am1_matches  # type: ignore[union-attr]
                    for index in key.atom_indices
                }

            if matched_atom_indices != expected_matches:
                # Handle the case where a handler could not fully assign the charges
                # to the whole molecule.
                continue

            matches.update(slot_matches if slot_matches is not None else {})
            matches.update(am1_matches if am1_matches is not None else {})  # type: ignore[arg-type]

            potentials.update(slot_potentials)
            potentials.update(am1_potentials)

            break

        found_matches = {index for key in matches for index in key.atom_indices}

        if found_matches != expected_matches:
            raise RuntimeError(
                f"{unique_molecule.to_smiles(explicit_hydrogens=False)} could "
                "not be fully assigned charges. Charges were assigned to atoms "
                f"{found_matches} but the molecule contains {expected_matches}.",
            )

        return matches, potentials

    @classmethod
    def _assign_charges_from_molecules(
        cls,
        unique_molecule: Molecule,
        molecules_with_preset_charges=list[Molecule] | None,
    ) -> tuple[bool, dict, dict]:
        if not molecules_with_preset_charges:
            return False, dict(), dict()

        for molecule_with_charges in molecules_with_preset_charges:
            if molecule_with_charges.is_isomorphic_with(unique_molecule):
                break
        else:
            return False, dict(), dict()

        _, atom_map = Molecule.are_isomorphic(
            molecule_with_charges,
            unique_molecule,
            return_atom_map=True,
        )

        matches = dict()
        potentials = dict()
        mapped_smiles = unique_molecule.to_smiles(mapped=True, explicit_hydrogens=True)

        for index_in_molecule_with_charges, partial_charge in enumerate(
            molecule_with_charges.partial_charges,
        ):
            index_in_topology = atom_map[index_in_molecule_with_charges]
            topology_key = SingleAtomChargeTopologyKey(
                this_atom_index=index_in_topology,
                extras={"handler": "preset"},
            )
            potential_key = PotentialKey(
                id=mapped_smiles,
                mult=index_in_molecule_with_charges,  # Not sure this prevents clashes in some corner cases
                associated_handler="molecules_with_preset_charges",
                bond_order=None,
            )
            potential = Potential(parameters={"charge": partial_charge})
            matches[topology_key] = potential_key
            potentials[potential_key] = potential

        return True, matches, potentials

    def store_matches(
        self,
        parameter_handler: ElectrostaticsHandlerType | list[ElectrostaticsHandlerType],
        topology: Topology,
        molecules_with_preset_charges=None,
        allow_nonintegral_charges: bool = False,
    ) -> None:
        """
        Populate self.key_map with key-val pairs of slots and unique potential identifiers.
        """
        # Reshape the parameter handlers into a dictionary for easier referencing.
        parameter_handlers = {
            handler.TAGNAME: handler
            for handler in (parameter_handler if isinstance(parameter_handler, list) else [parameter_handler])
        }

        self.potentials = dict()
        self.key_map = dict()

        groups = topology.identical_molecule_groups

        for unique_molecule_index, group in groups.items():
            unique_molecule = topology.molecule(unique_molecule_index)

            flag, matches, potentials = self._assign_charges_from_molecules(
                unique_molecule,
                molecules_with_preset_charges,
            )
            # TODO: Here is where the toolkit calls self.check_charges_assigned(). Do we skip this
            #       entirely given that we are not accepting `molecules_with_preset_charges`?

            if not flag:
                # TODO: Rename this method to something like `_find_matches`
                matches, potentials = self._find_reference_matches(
                    parameter_handlers,
                    unique_molecule,
                )

            self.potentials.update(potentials)

            for unique_molecule_atom in unique_molecule.atoms:
                unique_molecule_atom_index = unique_molecule.atom_index(
                    unique_molecule_atom,
                )

                for duplicate_molecule_index, atom_map in group:
                    duplicate_molecule = topology.molecule(duplicate_molecule_index)
                    duplicate_molecule_atom_index = atom_map[unique_molecule_atom_index]
                    duplicate_molecule_atom = duplicate_molecule.atom(
                        duplicate_molecule_atom_index,
                    )
                    topology_atom_index = topology.atom_index(duplicate_molecule_atom)

                    # Copy the keys associated with the reference molecule to the duplicate molecule
                    for key in matches:
                        if key.this_atom_index == unique_molecule_atom_index:
                            new_key = key.__class__(**key.model_dump())
                            new_key.this_atom_index = topology_atom_index

                            # Have this new key (on a duplicate molecule) point to the same potential
                            # as the old key (on a unique/reference molecule)

                            self.key_map[new_key] = matches[key]

            # don't bother going through logging machinery if it's not opted-in to
            if logger.isEnabledFor(logging.DEBUG):
                self._log_charge_provenance(unique_molecule, new_key)

        # this logic is all only to satisfy the allow_nonintegral_charges argument - could be more performant
        # to do it while assignment is happening
        if not allow_nonintegral_charges:
            topology_charges = [0.0] * topology.n_atoms

            for key, val in self.charges.items():
                topology_charges[key.atom_indices[0]] = val.m

            # TODO: Better data structures in Topology.identical_molecule_groups will make this
            #       cleaner and possibly more performant
            for molecule in topology.molecules:
                molecule_charges = [0.0] * molecule.n_atoms

                for atom in molecule.atoms:
                    molecule_index = molecule.atom_index(atom)
                    topology_index = topology.atom_index(atom)

                    molecule_charges[molecule_index] = topology_charges[topology_index]

                charge_sum = sum(molecule_charges)
                formal_sum = molecule.total_charge.m

                if abs(charge_sum - formal_sum) > 0.01:
                    raise NonIntegralMoleculeChargeError(
                        f"Molecule {molecule.to_smiles(explicit_hydrogens=False)} has "
                        f"a net charge of {charge_sum} compared to a total formal charge of "
                        f"{formal_sum}.",
                    )

    @staticmethod
    def _log_charge_provenance(
        unique_molecule: Molecule,
        key: ChargeModelTopologyKey | SingleAtomChargeTopologyKey | LibraryChargeTopologyKey,
    ):
        hill_formaula = unique_molecule.to_hill_formula()

        try:
            inchikey = unique_molecule.to_inchikey()
        except Exception as error:
            # In principle this is more like a warning, but the context is already logging.DEBUG, so why not?
            logger.debug(
                f"Could not generate InChIKey for molecule with Hill formula {hill_formaula}. "
                "This molecule's charge provenance will be logged without a valid InChIKey. "
                f"Error was {error}",
            )
            inchikey = "UNKNOWN_INCHIKEY"

        id_string = f"molecule with Hill formula {hill_formaula} and InChIKey {inchikey}"

        if type(key) is LibraryChargeTopologyKey:
            logger.debug(
                "Charge section LibraryCharges, applied to " + id_string,
            )

        elif type(key) is SingleAtomChargeTopologyKey:
            if key.extras["handler"] == "ToolkitAM1BCCHandler":
                logger.debug(
                    "Charge section ToolkitAM1BCC, using charge method "
                    f"{key.extras['partial_charge_method']}, "
                    f"applied to " + id_string,
                )
            elif key.extras["handler"] == "NAGLChargesHandler":
                logger.debug(
                    "Charge section NAGLCharges, using NAGL model "
                    f"{key.extras['partial_charge_method']}, "
                    f"applied to " + id_string,
                )

            elif key.extras["handler"] == "preset":
                logger.debug(
                    "Preset charges applied to " + id_string,
                )

            else:
                raise ValueError(f"Unhandled handler {key.extras['handler']}")

        elif type(key) is ChargeModelTopologyKey:
            logger.debug(
                "Charge section ChargeIncrementModel, using charge method "
                f"{key.partial_charge_method}, "
                f"applied to " + id_string,
            )

        elif type(key) is ChargeIncrementTopologyKey:
            # here is where the individual increments could be logged at creation time, but they're
            # also logged in _get_charges
            pass

        else:
            raise ValueError(f"Unhandled key type {type(key)}")

    def store_potentials(
        self,
        parameter_handler: ElectrostaticsHandlerType | list[ElectrostaticsHandlerType],
    ) -> None:
        """
        Populate self.potentials with key-val pairs of [PotentialKey, Potential].

        """
        # This logic is handled by ``store_matches`` as we may need to create potentials
        # to store depending on the handler type.



import itertools
import re
from collections import defaultdict

from openff.toolkit import Molecule, Quantity
from openff.toolkit.topology import Atom
from openff.toolkit.topology._mm_molecule import _SimpleMolecule
from openff.units.elements import MASSES, SYMBOLS

from openff.interchange import Interchange
from openff.interchange._annotations import PositiveFloat
from openff.interchange.components.potentials import Collection
from openff.interchange.components.toolkit import _get_14_pairs
from openff.interchange.exceptions import UnsupportedExportError
from openff.interchange.interop._virtual_sites import (
    _virtual_site_parent_molecule_mapping,
    get_positions_with_virtual_sites,
)
from openff.interchange.interop.common import _build_particle_map
from openff.interchange.interop.gromacs.export._virtual_sites import (
    _create_gromacs_virtual_site,
)
from openff.interchange.interop.gromacs.models.models import (
    GROMACSAngle,
    GROMACSAtom,
    GROMACSBond,
    GROMACSExclusion,
    GROMACSMolecule,
    GROMACSPair,
    GROMACSSettles,
    GROMACSSystem,
    GROMACSVirtualSite,
    LennardJonesAtomType,
    PeriodicImproperDihedral,
    PeriodicProperDihedral,
    RyckaertBellemansDihedral,
)
from openff.interchange.models import (
    BaseVirtualSiteKey,
    BondKey,
    LibraryChargeTopologyKey,
    SingleAtomChargeTopologyKey,
    TopologyKey,
)
from openff.interchange.smirnoff._virtual_sites import (
    _create_virtual_site_object,
)

type MoleculeLike = Molecule | _SimpleMolecule

_WATER = Molecule.from_mapped_smiles("[H:2][O:1][H:3]")
_SIMPLE_WATER = _SimpleMolecule.from_molecule(_WATER)


def _convert(
    interchange: Interchange,
    hydrogen_mass: PositiveFloat = 1.007947,
) -> GROMACSSystem:
    """Convert an `Interchange` object to `GROMACSSystem`."""
    if "vdW" in interchange.collections:
        nonbonded_function = 1
        scale_lj = interchange["vdW"].scale_14
        _combination_rule = interchange["vdW"].mixing_rule.lower()
        gen_pairs = True
    else:
        raise UnsupportedExportError(
            "Could not find a handler for short-ranged vdW interactions that is compatible with GROMACS.",
        )

    if _combination_rule == "lorentz-berthelot":
        combination_rule = 2
    elif _combination_rule == "geometric":
        combination_rule = 3
    else:
        raise UnsupportedExportError(
            f"Could not find a GROMACS-compatible combination rule for mixing rule {_combination_rule}.",
        )

    scale_electrostatics = interchange["Electrostatics"].scale_14

    system = GROMACSSystem(
        name="FOO",
        nonbonded_function=nonbonded_function,
        combination_rule=combination_rule,
        gen_pairs=gen_pairs,
        vdw_14=scale_lj,
        coul_14=scale_electrostatics,
    )

    unique_molecule_map: dict[
        int,
        list,
    ] = interchange.topology.identical_molecule_groups

    virtual_site_molecule_map: dict[
        BaseVirtualSiteKey,
        int,
    ] = _virtual_site_parent_molecule_mapping(interchange)

    molecule_virtual_site_map: dict[int, list[BaseVirtualSiteKey]] = defaultdict(list)

    for virtual_site_key, molecule_index in virtual_site_molecule_map.items():
        molecule_virtual_site_map[molecule_index].append(virtual_site_key)

    particle_map = _build_particle_map(
        interchange,
        molecule_virtual_site_map,
        collate=True,
    )

    # Give each atom in each unique molecule a unique name so that can act like an atom type

    _atom_atom_type_map: dict[Atom, str] = dict()

    try:
        vdw_collection = interchange["vdW"]
        interchange["Electrostatics"]
    except KeyError:
        raise UnsupportedExportError("Plugins not implemented.")

    for unique_molecule_index in unique_molecule_map:
        unique_molecule = interchange.topology.molecule(unique_molecule_index)

        # If this molecule doesn't have a name ("^$", empty string), name it MOL0 incrementing
        # Also rename it if it's already MOL\d+ since that was probably assigned by this function
        # earlier in a pipeline. The molecule_types dict keys by molecule names and it's important
        # that they are unique (in the same way that the unitand moleucle names are)
        if re.match(
            r"^$|MOL\d+",
            getattr(unique_molecule, "name", ""),  # SimpleMolecule might not have .name
        ):
            unique_molecule.name = "MOL" + str(unique_molecule_index)

        if unique_molecule.name in system.molecule_types:
            raise RuntimeError(
                "Problem keeping molecule names unique. This should not be possible - please"
                "raise an issue describing how this error occurred.",
            )

        for atom in unique_molecule.atoms:
            atom_type_name = f"{unique_molecule.name}_{particle_map[unique_molecule.atom_index(atom)]}"
            _atom_atom_type_map[atom] = atom_type_name

            # when looking up parameters, use the topology index, not the particle index ...
            # ... or so I think is the expectation of the `TopologyKey`s in the vdW collection
            topology_index = interchange.topology.atom_index(atom)
            key: TopologyKey | BaseVirtualSiteKey | LibraryChargeTopologyKey = TopologyKey(
                atom_indices=(topology_index,),
            )

            vdw_parameters = vdw_collection.potentials[vdw_collection.key_map[key]].parameters

            # Build atom types
            system.atom_types[atom_type_name] = LennardJonesAtomType(
                name=_atom_atom_type_map[atom],
                bonding_type="",
                atomic_number=atom.atomic_number,
                mass=MASSES[atom.atomic_number],
                charge=Quantity(0.0, "elementary_charge"),
                particle_type="A",
                sigma=vdw_parameters["sigma"].to("nanometer"),
                epsilon=vdw_parameters["epsilon"].to("kilojoule_per_mole"),
            )

        for virtual_site_key in molecule_virtual_site_map[interchange.topology.molecule_index(unique_molecule)]:
            atom_type_name = f"{unique_molecule.name}_{particle_map[virtual_site_key]}"
            _atom_atom_type_map[virtual_site_key] = atom_type_name

            topology_index = particle_map[virtual_site_key]

            vdw_parameters = vdw_collection.potentials[vdw_collection.key_map[virtual_site_key]].parameters

            # TODO: Separate class for "atom types" representing virtual sites?
            system.atom_types[atom_type_name] = LennardJonesAtomType(
                name=_atom_atom_type_map[virtual_site_key],
                bonding_type="",
                atomic_number=0,
                mass=Quantity(0.0, "dalton"),
                charge=Quantity(0.0, "elementary_charge"),
                particle_type="D",
                sigma=vdw_parameters["sigma"].to("nanometer"),
                epsilon=vdw_parameters["epsilon"].to("kilojoule_per_mole"),
            )

    _partial_charges: dict[int | BaseVirtualSiteKey, float] = dict()

    # Indexed by particle (atom or virtual site) indices
    for key, charge in interchange["Electrostatics"].charges.items():
        if type(key) is TopologyKey:
            _partial_charges[key.atom_indices[0]] = charge
        elif isinstance(key, BaseVirtualSiteKey):
            _partial_charges[key] = charge
        elif type(key) is SingleAtomChargeTopologyKey:
            _partial_charges[key.this_atom_index] = charge
        else:
            raise RuntimeError(f"Unexpected type {type(key)=}")

    _ordered_molecules = [None] * interchange.topology.n_molecules

    for unique_molecule_index in unique_molecule_map:
        unique_molecule = interchange.topology.molecule(unique_molecule_index)

        if unique_molecule.name == "":
            unique_molecule.name = "MOL" + str(unique_molecule_index)

        molecule = GROMACSMolecule(name=unique_molecule.name)

        unique_residue_names = {atom.metadata.get("residue_name", None) for atom in unique_molecule.atoms}

        if None in unique_residue_names:
            if len(unique_residue_names) > 1:
                raise NotImplementedError(
                    "If some atoms have residue names, all atoms must have residue names.",
                )
            else:
                # Use dummy since we're already iterating over this molecule's atoms
                for _atom in unique_molecule.atoms:
                    _atom.metadata["residue_name"] = unique_molecule.name

        for atom in unique_molecule.atoms:
            name = SYMBOLS[atom.atomic_number] if getattr(atom, "name", "") == "" else atom.name

            charge = _partial_charges[interchange.topology.atom_index(atom)]

            molecule.atoms.append(
                GROMACSAtom(
                    index=unique_molecule.atom_index(atom) + 1,
                    name=name,
                    atom_type=_atom_atom_type_map[atom],
                    residue_index=atom.metadata.get(
                        "residue_number",
                        unique_molecule_index + 1,
                    ),
                    residue_name=atom.metadata["residue_name"],
                    charge_group_number=1,
                    charge=charge,
                    mass=MASSES[atom.atomic_number],
                ),
            )

        this_molecule_atom_type_names = tuple(atom.atom_type for atom in molecule.atoms)

        molecule._contained_atom_types = {
            atom_type_name: system.atom_types[atom_type_name] for atom_type_name in this_molecule_atom_type_names
        }

        # Use a set to de-duplicate
        pairs: set[tuple] = {*_get_14_pairs(unique_molecule)}

        for pair in pairs:
            molecule_indices = sorted(unique_molecule.atom_index(atom) for atom in pair)

            if system.gen_pairs:
                molecule.pairs.append(
                    GROMACSPair(
                        atom1=molecule_indices[0] + 1,
                        atom2=molecule_indices[1] + 1,
                    ),
                )

            else:
                raise NotImplementedError()

        _convert_settles(molecule, unique_molecule, interchange)
        _convert_bonds(molecule, unique_molecule, interchange)
        _convert_angles(molecule, unique_molecule, interchange)
        # pairs
        _convert_dihedrals(molecule, unique_molecule, interchange)
        # other constraints?

        _convert_virtual_sites(
            molecule,
            unique_molecule,
            interchange,
            molecule_virtual_site_map,
            _atom_atom_type_map,
        )

        # Apply HMR to this molecule only
        _apply_hmr(molecule, unique_molecule, hydrogen_mass)

        system.molecule_types[unique_molecule.name] = molecule

        for duplicate_molecule_index, _ in unique_molecule_map[unique_molecule_index]:
            _ordered_molecules[duplicate_molecule_index] = unique_molecule.name

    if "VirtualSites" in interchange.collections:
        # TODO: Some say to skip this if the user only wants a topology file?

        system.positions = get_positions_with_virtual_sites(
            interchange,
            collate=True,
        )

    else:
        system.positions = interchange.positions

    system.molecules = [(key, len(list(group))) for key, group in itertools.groupby(_ordered_molecules)]  # type: ignore[misc]

    system.box = interchange.box

    return system


def _convert_bonds(
    molecule: GROMACSMolecule,
    unique_molecule: MoleculeLike,
    interchange: Interchange,
):
    if len(molecule.settles) > 0:
        return

    try:
        collection = interchange["Bonds"]
    except LookupError:
        return

    # if this is slow, pass it among valence converters
    atom_indices_in_this_molecule = {interchange.topology.atom_index(a) for a in unique_molecule.atoms}

    offset = min(atom_indices_in_this_molecule)

    for top_key in collection.key_map:
        if top_key.atom_indices[0] not in atom_indices_in_this_molecule:
            continue

        molecule.bonds.append(
            _create_single_bond(
                top_key,
                collection,
                offset,
            ),
        )


def _create_single_bond(
    top_key,
    collection,
    offset: int,
) -> GROMACSBond:
    params = collection.potentials[collection.key_map[top_key]].parameters

    return GROMACSBond(
        atom1=top_key.atom_indices[0] - offset + 1,
        atom2=top_key.atom_indices[1] - offset + 1,
        function=1,
        length=params["length"].to("nanometer"),
        k=params["k"].to("kilojoule_per_mole / nanometer**2"),
    )


def _convert_angles(
    molecule: GROMACSMolecule,
    unique_molecule: MoleculeLike,
    interchange: Interchange,
):
    if len(molecule.settles) > 0:
        return

    try:
        collection = interchange["Angles"]
    except LookupError:
        return

    # If this is slow, pass it among valence converters
    atom_indices_in_this_molecule = {interchange.topology.atom_index(a) for a in unique_molecule.atoms}

    offset = min(atom_indices_in_this_molecule)

    for top_key in collection.key_map:
        if top_key.atom_indices[0] not in atom_indices_in_this_molecule:
            continue

        molecule.angles.append(
            _create_single_angle(
                top_key,
                collection,
                offset,
            ),
        )


def _create_single_angle(
    top_key,
    collection,
    offset: int,
) -> GROMACSAngle:
    params = collection.potentials[collection.key_map[top_key]].parameters

    return GROMACSAngle(
        atom1=top_key.atom_indices[0] - offset + 1,
        atom2=top_key.atom_indices[1] - offset + 1,
        atom3=top_key.atom_indices[2] - offset + 1,
        angle=params["angle"].to("degree"),
        k=params["k"].to("kilojoule_per_mole / radian**2"),
    )


def _convert_dihedrals(
    molecule: GROMACSMolecule,
    unique_molecule: MoleculeLike,
    interchange: Interchange,
):
    rb_torsion_handler: Collection | None = interchange.collections.get(
        "RBTorsions",
        None,
    )
    proper_torsion_handler: Collection | None = interchange.collections.get(
        "ProperTorsions",
        None,
    )
    improper_torsion_handler: Collection | None = interchange.collections.get(
        "ImproperTorsions",
        None,
    )

    atom_indices_in_this_molecule = {interchange.topology.atom_index(a) for a in unique_molecule.atoms}

    offset = min(atom_indices_in_this_molecule)

    if proper_torsion_handler:
        for top_key in proper_torsion_handler.key_map:
            # assume that all atoms in this torsion are in the same molecule,
            # so if the index of the first atom in the torsion is not in the unique
            # molecule, skip the molecule altogether
            if top_key.atom_indices[0] not in atom_indices_in_this_molecule:
                continue

            molecule.dihedrals.append(
                _create_single_dihedral(
                    top_key,
                    proper_torsion_handler,
                    offset,
                ),
            )

    if rb_torsion_handler:
        for top_key in rb_torsion_handler.key_map:
            if top_key.atom_indices[0] not in atom_indices_in_this_molecule:
                continue

            molecule.dihedrals.append(
                _create_single_rb_torsion(
                    top_key,
                    rb_torsion_handler,
                    offset,
                ),
            )

    # TODO: Ensure number of torsions written matches what is expected
    if improper_torsion_handler:
        # Molecule/Topology.impropers lists the central atom **second** ...
        for improper in unique_molecule.smirnoff_impropers:
            topology_indices = tuple(interchange.topology.atom_index(a) for a in improper)
            # ... so the tuple must be modified to list the central atom **first**,
            # which is how the improper handler's slot map is built up
            indices_to_match = (
                topology_indices[1],
                topology_indices[0],
                topology_indices[2],
                topology_indices[3],
            )

            molecule_indices = tuple(unique_molecule.atom_index(a) for a in improper)

            # Now, indices_to_match has the central atom listed **first**,
            # but it's still listed second in molecule_indices

            for top_key in improper_torsion_handler.key_map:
                if top_key.atom_indices[0] != indices_to_match[0]:
                    continue
                if top_key.atom_indices[1] != indices_to_match[1]:
                    continue
                if top_key.atom_indices[2] != indices_to_match[2]:
                    continue
                if top_key.atom_indices[3] != indices_to_match[3]:
                    continue
                if indices_to_match == top_key.atom_indices:
                    key = improper_torsion_handler.key_map[top_key]
                    params = improper_torsion_handler.potentials[key].parameters

                    idivf = int(params["idivf"])

                    molecule.dihedrals.append(
                        PeriodicImproperDihedral(
                            atom1=molecule_indices[1] + 1,
                            atom2=molecule_indices[0] + 1,
                            atom3=molecule_indices[2] + 1,
                            atom4=molecule_indices[3] + 1,
                            phi=params["phase"].to("degree"),
                            k=params["k"].to("kilojoule_per_mole") / idivf,
                            multiplicity=int(params["periodicity"]),
                        ),
                    )


def _create_single_dihedral(
    top_key,
    proper_torsion_handler,
    offset: int,
) -> PeriodicProperDihedral:
    params = proper_torsion_handler.potentials[proper_torsion_handler.key_map[top_key]].parameters

    # skip dimensionality check, trust it's dimensionless
    idivf = int(params["idivf"].m) if "idivf" in params else 1

    return PeriodicProperDihedral(
        atom1=top_key.atom_indices[0] - offset + 1,
        atom2=top_key.atom_indices[1] - offset + 1,
        atom3=top_key.atom_indices[2] - offset + 1,
        atom4=top_key.atom_indices[3] - offset + 1,
        phi=params["phase"].to("degree"),
        k=params["k"].to("kilojoule_per_mole") / idivf,
        multiplicity=int(
            params["periodicity"].m,
        ),  # skip  dimension check, trust it's demensionless
    )


def _create_single_rb_torsion(
    top_key,
    rb_torsion_handler,
    offset: int,
) -> RyckaertBellemansDihedral:
    params = rb_torsion_handler.potentials[rb_torsion_handler.key_map[top_key]].parameters

    return RyckaertBellemansDihedral(
        atom1=top_key.atom_indices[0] - offset + 1,
        atom2=top_key.atom_indices[1] - offset + 1,
        atom3=top_key.atom_indices[2] - offset + 1,
        atom4=top_key.atom_indices[3] - offset + 1,
        c0=params["c0"],
        c1=params["c1"],
        c2=params["c2"],
        c3=params["c3"],
        c4=params["c4"],
        c5=params["c5"],
    )


def _convert_virtual_sites(
    molecule: GROMACSMolecule,
    unique_molecule: MoleculeLike,
    interchange: Interchange,
    molecule_virtual_site_map,
    _atom_atom_type_map,
):
    if "VirtualSites" not in interchange.collections:
        return

    if len(molecule_virtual_site_map) == 0:
        return

    for virtual_site_key in molecule_virtual_site_map[interchange.topology.molecule_index(unique_molecule)]:
        virtual_site_potential = interchange["VirtualSites"].potentials[
            interchange["VirtualSites"].key_map[virtual_site_key]
        ]

        virtual_site_object = _create_virtual_site_object(
            virtual_site_key,
            virtual_site_potential,
        )

        # TODO: Store this somewhere so it doesn't need to be computed twice
        particle_map = _build_particle_map(
            interchange,
            molecule_virtual_site_map,
            collate=True,
        )

        gromacs_virtual_site: GROMACSVirtualSite = _create_gromacs_virtual_site(
            interchange,
            virtual_site_object,
            virtual_site_key,
            particle_map,
        )

        molecule.virtual_sites.append(gromacs_virtual_site)

        # TODO: Consider making a separate type for this non-atom "atom"
        molecule.atoms.append(
            GROMACSAtom(
                index=gromacs_virtual_site.site,
                name=gromacs_virtual_site.name,
                atom_type=_atom_atom_type_map[virtual_site_key],
                residue_index=molecule.atoms[0].residue_index,
                residue_name=molecule.atoms[0].residue_name,
                charge_group_number=1,
                charge=interchange["Electrostatics"].charges[virtual_site_key],
                mass=Quantity(0.0, "dalton"),
            ),
        )

        molecule.exclusions.append(
            GROMACSExclusion(
                first_atom=gromacs_virtual_site.site,
                other_atoms=gromacs_virtual_site.orientation_atoms,
            ),
        )

        # Exclusions need to be listed twice, i.e. i-j and then j-i
        # not obvious how orientation atom-orientation atom exclusions are to be treated here;
        # for now just add the symmetries between virtual site and orientation atoms

        for orientation_atom in gromacs_virtual_site.orientation_atoms:
            molecule.exclusions.append(
                GROMACSExclusion(
                    first_atom=orientation_atom,
                    other_atoms=[gromacs_virtual_site.site],
                ),
            )


def _convert_settles(
    molecule: GROMACSMolecule,
    unique_molecule: MoleculeLike,
    interchange: Interchange,
):
    if "Constraints" not in interchange.collections:
        return

    if isinstance(unique_molecule, Molecule):
        if not unique_molecule.is_isomorphic_with(_WATER):
            return
    elif isinstance(unique_molecule, _SimpleMolecule):
        if not unique_molecule.is_isomorphic_with(_SIMPLE_WATER):
            return

    if unique_molecule.atom(0).atomic_number != 8:
        raise Exception(
            "Writing `[ settles ]` assumes water is ordered as OHH. Please raise an issue "
            "if you would benefit from this assumption changing.",
        )

    topology_atom_indices = [interchange.topology.atom_index(atom) for atom in unique_molecule.atoms]

    constraint_lengths = set()

    for atom_pair in itertools.combinations(topology_atom_indices, 2):
        key = BondKey(atom_indices=atom_pair)

        # First grab SETTLES distances from constraints, not bond lengths
        if key not in interchange["Constraints"].key_map:
            return

        constraints = interchange["Constraints"]

        try:
            constraint_lengths.add(
                constraints.potentials[constraints.key_map[key]].parameters["distance"],
            )
        except LookupError:
            try:
                bonds = interchange["Bonds"]
                constraint_lengths.add(
                    bonds.potentials[bonds.key_map[key]].parameters["length"],
                )
            except LookupError:
                raise RuntimeError(
                    f"Could not find a constraint distance for atoms {key.atom_indices=}",
                )

    if len(constraint_lengths) != 2:
        raise RuntimeError(
            "Found unexpected number of unique constraint lengths in constrained water.",
        )

    molecule.settles.append(
        GROMACSSettles(
            first_atom=1,  # TODO: documentation unclear on if this is first or oxygen
            oxygen_hydrogen_distance=min(constraint_lengths),
            hydrogen_hydrogen_distance=max(constraint_lengths),
        ),
    )

    # Exclusions need to be listed twice, i.e. i-j and then j-i
    molecule.exclusions.append(
        GROMACSExclusion(
            first_atom=1,
            other_atoms=[2, 3],
        ),
    )
    molecule.exclusions.append(
        GROMACSExclusion(
            first_atom=2,
            other_atoms=[1, 3],
        ),
    )
    molecule.exclusions.append(
        GROMACSExclusion(
            first_atom=3,
            other_atoms=[1, 2],
        ),
    )


# TODO: Refactor this with the OpenMM version into a common mass mask?
def _apply_hmr(
    gromacs_molecule: GROMACSMolecule,
    toolkit_molecule: Molecule,
    hydrogen_mass: PositiveFloat,
):
    if abs(hydrogen_mass - 1.008) < 1e-3:
        return

    if len(gromacs_molecule.virtual_sites) > 0:
        raise UnsupportedExportError(
            "Hydrogen mass repartitioning with virtual sites present, even on rigid water, is not yet supported.",
        )

    water = Molecule.from_smiles("O")

    def _is_water(molecule: Molecule) -> bool:
        return molecule.is_isomorphic_with(water)

    _hydrogen_mass = Quantity(hydrogen_mass, "dalton")

    for bond in toolkit_molecule.bonds:
        heavy_atom, hydrogen_atom = bond.atoms

        if heavy_atom.atomic_number == 1:
            heavy_atom, hydrogen_atom = hydrogen_atom, heavy_atom

        # TODO: This should only skip rigid waters, even though HMR or flexible water is questionable
        if (
            (hydrogen_atom.atomic_number == 1)
            and (heavy_atom.atomic_number != 1)
            and not (_is_water(hydrogen_atom.molecule))
        ):
            # these are molecule indices, whereas in the OpenMM function they are topology indices
            # these are indexed to the toolkit molecule (0-index), not the GROMACS molecule (1-index),
            # although indexing into GROMACSMolecule.atoms is 0-indexed, GROAMCSAtom.index is 1-indexed
            hydrogen_index = hydrogen_atom.molecule_atom_index
            heavy_index = heavy_atom.molecule_atom_index

            mass_to_transfer = _hydrogen_mass - gromacs_molecule.atoms[hydrogen_index].mass

            gromacs_molecule.atoms[hydrogen_index].mass += mass_to_transfer
            gromacs_molecule.atoms[heavy_index].mass -= mass_to_transfer



from typing import Any

class Forcefield:
    def __init__(self, name: int): ...
    def get_parameters(
        self,
        group: str,
        key: str | list[str],
        keys_are_atom_classes: bool = False,
    ) -> dict[str, float]: ...
    @staticmethod
    def get_generator(ff: Forcefield, gen_type: Any) -> Any: ...
    @property
    def lj14scale(self) -> float: ...
    @property
    def coulomb14scale(self) -> float: ...



from abc import abstractmethod
from copy import copy
from typing import TYPE_CHECKING

from openff.toolkit import Quantity, Topology

from openff.interchange.components.potentials import Collection, Potential
from openff.interchange.foyer._guard import has_foyer
from openff.interchange.models import PotentialKey, TopologyKey

if TYPE_CHECKING:
    if has_foyer:
        try:
            from foyer import Forcefield
        except ModuleNotFoundError:
            pass


# Is this the safest way to achieve PotentialKey id separation?
POTENTIAL_KEY_SEPARATOR = "-"


def _copy_params(
    params: dict[str, float],
    *drop_keys: str,
    param_units: dict[str, str] | None = None,
) -> dict:
    """Copy parameters from a dictionary."""
    params_copy = copy(params)
    for drop_key in drop_keys:
        params_copy.pop(drop_key, None)
    if param_units:
        for unit_item, unit_str in param_units.items():
            params_copy[unit_item] = Quantity(params_copy[unit_item], unit_str)
    return params_copy


def _get_potential_key_id(atom_slots: dict[TopologyKey, PotentialKey], idx):
    """From a dictionary of TopologyKey: PotentialKey, get the PotentialKey id."""
    top_key = TopologyKey(atom_indices=(idx,))
    return atom_slots[top_key].id


class FoyerConnectedAtomsHandler(Collection):
    """Base class for handlers storing valence potentials produced by a Foyer force field."""

    force_field_key: str
    connection_attribute: str = ""
    raise_on_missing_params: bool = True

    def store_matches(
        self,
        atom_slots: dict[TopologyKey, PotentialKey],
        topology: "Topology",
    ) -> None:
        """Populate self.key_map with key-val pairs of [TopologyKey, PotentialKey]."""
        for connection in getattr(topology, self.connection_attribute):
            try:
                atoms_iterable = connection.atoms
            except AttributeError:
                atoms_iterable = connection
            atom_indices = tuple(topology.atom_index(atom) for atom in atoms_iterable)

            top_key = TopologyKey(atom_indices=atom_indices)
            pot_key_ids = tuple(_get_potential_key_id(atom_slots, idx) for idx in atom_indices)

            self.key_map[top_key] = PotentialKey(
                id=POTENTIAL_KEY_SEPARATOR.join(pot_key_ids),
            )

    def store_potentials(self, force_field: "Forcefield") -> None:
        """Populate self.potentials with key-val pairs of [PotentialKey, Potential]."""
        from foyer.exceptions import MissingForceError, MissingParametersError

        for pot_key in self.key_map.values():
            try:
                params = force_field.get_parameters(
                    self.force_field_key,
                    key=pot_key.id.split(POTENTIAL_KEY_SEPARATOR),
                )
                params = self.get_params_with_units(params)
                self.potentials[pot_key] = Potential(parameters=params)
            except MissingForceError:
                # Here, we can safely assume that the ForceGenerator is Missing
                self.key_map = {}
                self.potentials = {}
                return
            except MissingParametersError as e:
                if self.raise_on_missing_params:
                    raise e
                else:
                    pass

    @abstractmethod
    def get_params_with_units(self, params):
        """Get the parameters of this handler, tagged with units."""
        raise NotImplementedError



"""The interface between Interchange and SMIRNOFF objects."""

from openff.interchange.smirnoff._base import SMIRNOFFCollection
from openff.interchange.smirnoff._gbsa import SMIRNOFFGBSACollection
from openff.interchange.smirnoff._nonbonded import (
    SMIRNOFFElectrostaticsCollection,
    SMIRNOFFvdWCollection,
)
from openff.interchange.smirnoff._valence import (
    SMIRNOFFAngleCollection,
    SMIRNOFFBondCollection,
    SMIRNOFFConstraintCollection,
    SMIRNOFFImproperTorsionCollection,
    SMIRNOFFProperTorsionCollection,
)
from openff.interchange.smirnoff._virtual_sites import SMIRNOFFVirtualSiteCollection

__all__ = (
    "SMIRNOFFAngleCollection",
    "SMIRNOFFBondCollection",
    "SMIRNOFFCollection",
    "SMIRNOFFConstraintCollection",
    "SMIRNOFFElectrostaticsCollection",
    "SMIRNOFFGBSACollection",
    "SMIRNOFFImproperTorsionCollection",
    "SMIRNOFFProperTorsionCollection",
    "SMIRNOFFVirtualSiteCollection",
    "SMIRNOFFvdWCollection",
)
